import re
from os import environ
from pathlib import Path
from sys import exit

from epstein_files.util.constant.strings import FILE_NAME_REGEX, FILE_STEM_REGEX, HOUSE_OVERSIGHT_PREFIX

EPSTEIN_DOCS_DIR_ENV_VAR_NAME = 'EPSTEIN_DOCS_DIR'
DOCS_DIR_ENV = environ[EPSTEIN_DOCS_DIR_ENV_VAR_NAME]
DOCS_DIR = Path(DOCS_DIR_ENV or '').resolve()
PICKLED_PATH = Path("the_epstein_files.pkl.gz")

if not DOCS_DIR_ENV:
    print(f"ERROR: {EPSTEIN_DOCS_DIR_ENV_VAR_NAME} env var not set!")
    exit(1)
elif not DOCS_DIR.exists():
    print(f"ERROR: {EPSTEIN_DOCS_DIR_ENV_VAR_NAME}='{DOCS_DIR}' does not exist!")
    exit(1)

HTML_DIR = Path('docs')
EXTRACTED_EMAILS_DIR = Path('emails_extracted_from_legal_filings')
EPSTEIN_WORD_COUNT_HTML_PATH = HTML_DIR.joinpath('epstein_texts_and_emails_word_count.html')
GH_PAGES_HTML_PATH = HTML_DIR.joinpath('index.html')
JSON_METADATA_PATH = HTML_DIR.joinpath('epstein_files_nov_2025_cryptadamus_metadata.json')
WORD_COUNT_HTML_PATH = HTML_DIR.joinpath('epstein_emails_word_count.html')

BUILD_ARTIFACTS = [
    EPSTEIN_WORD_COUNT_HTML_PATH,
    GH_PAGES_HTML_PATH,
    JSON_METADATA_PATH,
    WORD_COUNT_HTML_PATH,
]

FILE_ID_REGEX = re.compile(fr".*{FILE_NAME_REGEX.pattern}")
FILENAME_LENGTH = len(HOUSE_OVERSIGHT_PREFIX) + 6
KB = 1024
MB = KB * KB


# Handles both string and int 'id' args.
id_str = lambda id: f"{int(id):06d}"
filename_for_id = lambda id: file_stem_for_id(id) + '.txt'


def coerce_file_stem(filename_or_id: int | str) -> str:
    """Generate a valid file_stem no matter what form the argument comes in."""
    if isinstance(filename_or_id, str) and filename_or_id.startswith(HOUSE_OVERSIGHT_PREFIX):
        file_id = extract_file_id(filename_or_id)
        file_stem = file_stem_for_id(file_id)
    else:
        file_stem = file_stem_for_id(filename_or_id)

    if not FILE_STEM_REGEX.match(file_stem):
        raise RuntimeError(f"Invalid stem '{file_stem}' from '{filename_or_id}'")

    return file_stem


def coerce_file_name(filename_or_id: int | str) -> str:
    return coerce_file_stem(filename_or_id) + '.txt'


def coerce_file_path(filename_or_id: int | str) -> Path:
    return DOCS_DIR.joinpath(coerce_file_name(filename_or_id))


def extract_file_id(filename_or_id: int | str | Path) -> str:
    if isinstance(filename_or_id, int) or (isinstance(filename_or_id, str) and len(filename_or_id) <= 6):
        return id_str(filename_or_id)

    file_match = FILE_ID_REGEX.match(str(filename_or_id))

    if not file_match:
        raise RuntimeError(f"Failed to extract file ID from {filename_or_id}")

    return file_match.group(1)


def file_size(file_path: str | Path) -> int:
    return Path(file_path).stat().st_size


def file_size_str(file_path: str | Path) -> str:
    size = file_size(file_path)
    digits = 2

    if size > MB:
        size_num = float(size) / MB
        size_str = 'MB'
    elif size > KB:
        size_num = float(size) / KB
        size_str = 'kb'
        digits = 1
    else:
        return f"{size} b"

    return f"{size_num:,.{digits}f} {size_str}"


def file_stem_for_id(id: int | str) -> str:
    if isinstance(id, int) or (isinstance(id, str) and len(id) <= 6):
        return f"{HOUSE_OVERSIGHT_PREFIX}{id_str(id)}"
    elif len(id) == 8:
        return f"{HOUSE_OVERSIGHT_PREFIX}{id}"
    else:
        raise RuntimeError(f"Unknown kind of file id {id}")


def is_local_extract_file(filename) -> bool:
    """Return true if filename is of form 'HOUSE_OVERSIGHT_029835_1.txt'."""
    file_match = FILE_ID_REGEX.match(str(filename))
    return True if file_match and file_match.group(2) else False


def make_clean() -> None:
    """Delete all build artifacts."""
    for build_file in BUILD_ARTIFACTS:
        if build_file.exists():
            print(f"Removing build file '{build_file}'...")
            build_file.unlink()



#!/usr/bin/env python
"""
Reformat Epstein text message files for readability and count email senders.
For use with iMessage log files from https://drive.google.com/drive/folders/1hTNH5woIRio578onLGElkTWofUSWRoH_

Install: 'poetry install'
    Run: 'EPSTEIN_DOCS_DIR=/path/to/TXT ./generate.py'
"""
from sys import exit

from dotenv import load_dotenv
load_dotenv()
from rich.padding import Padding

from epstein_files.documents.email import Email
from epstein_files.documents.messenger_log import  MessengerLog
from epstein_files.epstein_files import EpsteinFiles, count_by_month
from epstein_files.util.constant.html import *
from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import EMAIL_CLASS, MESSENGER_LOG_CLASS
from epstein_files.util.data import dict_sets_to_lists
from epstein_files.util.env import args, specified_names
from epstein_files.util.file_helper import GH_PAGES_HTML_PATH, JSON_METADATA_PATH, make_clean
from epstein_files.util.logging import logger
from epstein_files.util.rich import *
from epstein_files.util.timer import Timer

PRINT_COLOR_KEY_EVERY_N_EMAILS = 150

# Order matters. Default names to print emails for.
DEFAULT_EMAILERS = [
    JEREMY_RUBIN,
    AL_SECKEL,
    JOI_ITO,
    JABOR_Y,
    STEVEN_SINOFSKY,
    DANIEL_SIAD,
    JEAN_LUC_BRUNEL,
    STEVEN_HOFFENBERG,
    EHUD_BARAK,
    MARTIN_NOWAK,
    MASHA_DROKOVA,
    RENATA_BOLOTOVA,
    STEVE_BANNON,
    OLIVIER_COLOM,
    BORIS_NIKOLIC,
    PRINCE_ANDREW,
    JIDE_ZEITLIN,
    DAVID_STERN,
    MOHAMED_WAHEED_HASSAN,
    JENNIFER_JACQUET,
    None,
]

# Order matters. Default names to print tables w/email subject, timestamp, etc for.
# TODO: get rid of this
DEFAULT_EMAILER_TABLES: list[str | None] = [
    GHISLAINE_MAXWELL,
    LEON_BLACK,
    LANDON_THOMAS,
    KATHRYN_RUEMMLER,
    DARREN_INDYKE,
    RICHARD_KAHN,
    TYLER_SHEARS,
    SULTAN_BIN_SULAYEM,
    DEEPAK_CHOPRA,
    ARIANE_DE_ROTHSCHILD,
    TOM_PRITZKER,
]

if len(set(DEFAULT_EMAILERS).intersection(set(DEFAULT_EMAILER_TABLES))) > 0:
    raise RuntimeError(f"Some names appear in both DEFAULT_EMAILERS and DEFAULT_EMAILER_TABLES")


def generate_html() -> None:
    if args.make_clean:
        make_clean()
        exit()

    timer = Timer()
    epstein_files = EpsteinFiles.get_files(timer)

    if args.json_metadata:
        json_str = epstein_files.json_metadata()

        if args.build:
            with open(JSON_METADATA_PATH, 'w') as f:
                f.write(json_str)
                timer.print_at_checkpoint(f"Wrote {file_size_str(JSON_METADATA_PATH)} to '{JSON_METADATA_PATH}'")
        else:
            console.print_json(json_str, indent=4, sort_keys=True)

        exit()

    print_header(epstein_files)

    if args.colors_only:
        exit()

    if args.output_texts:
        _print_text_messages(epstein_files)
        timer.print_at_checkpoint(f'Printed {len(epstein_files.imessage_logs)} text message logs')

    if args.output_emails:
        emails_printed = _print_emails(epstein_files)
        timer.print_at_checkpoint(f"Printed {emails_printed:,} emails")

    if args.output_other_files:
        files_printed = epstein_files.print_other_files_table()
        timer.print_at_checkpoint(f"Printed {len(files_printed)} other files")

    # Save output
    write_html(GH_PAGES_HTML_PATH)
    logger.warning(f"Total time: {timer.seconds_since_start_str()}")

    # JSON stats (mostly used for building pytest checks)
    if args.json_stats:
        console.line(5)
        _print_json_stats(epstein_files)


def _print_emails(epstein_files: EpsteinFiles) -> int:
    """Returns number of emails printed."""
    print_section_header(('Selections from ' if not args.all_emails else '') + 'His Emails')
    print_other_site_link(is_header=False)

    if len(specified_names) == 0:
        epstein_files.print_emailer_counts_table()

    emailers_to_print: list[str | None]
    emailer_tables: list[str | None] = []
    already_printed_emails: list[Email] = []
    num_emails_printed_since_last_color_key = 0

    if args.all_emails:
        console.print('Email conversations are sorted chronologically based on time of the first email.')
        emailers_to_print = sorted(epstein_files.all_emailers(), key=lambda e: epstein_files.earliest_email_at(e))
        print_numbered_list_of_emailers(emailers_to_print, epstein_files)
    else:
        emailers_to_print = specified_names if specified_names else DEFAULT_EMAILERS
        console.print('Email conversations grouped by counterparty can be found in the order listed below.')
        print_numbered_list_of_emailers(emailers_to_print)
        console.print("\nAfter that there's tables linking to (but not displaying) all known emails for each of these people:")

        if len(specified_names) > 0:
            print_numbered_list_of_emailers(DEFAULT_EMAILER_TABLES)

    for author in emailers_to_print:
        newly_printed_emails = epstein_files.print_emails_for(author)
        already_printed_emails.extend(newly_printed_emails)
        num_emails_printed_since_last_color_key += len(newly_printed_emails)

        # Print color key every once in a while
        if num_emails_printed_since_last_color_key > PRINT_COLOR_KEY_EVERY_N_EMAILS:
            print_color_key()
            num_emails_printed_since_last_color_key = 0

    if not specified_names:
        if not args.all_emails:
            print_author_header(f"Email Tables for {len(emailer_tables)} Other People", 'white')

            for name in DEFAULT_EMAILER_TABLES:
                epstein_files.print_emails_table_for(name)

        epstein_files.print_email_device_info()

    # Check that all emails were actually printed
    if args.all_emails:
        email_ids_that_were_printed = set([email.file_id for email in already_printed_emails])
        logger.warning(f"Printed {len(already_printed_emails)} emails of {len(email_ids_that_were_printed)} unique file IDs.")

        for email in epstein_files.emails:
            if email.file_id not in email_ids_that_were_printed and not email.is_duplicate:
                logger.warning(f"Failed to print {email.summary()}")

    logger.warning(f"Rewrote {len(Email.rewritten_header_ids)} headers of {len(epstein_files.emails)} emails")
    return len(already_printed_emails)


def _print_text_messages(epstein_files: EpsteinFiles) -> None:
    print_section_header('Text Messages')
    print_centered("(conversations are sorted chronologically based on timestamp of first message)\n", style='gray30')
    authors: list[str | None] = specified_names if specified_names else [JEFFREY_EPSTEIN]
    log_files = epstein_files.imessage_logs_for(authors)

    for log_file in log_files:
        console.print(Padding(log_file))
        console.line(2)

    epstein_files.print_imessage_summary()


def _print_json_stats(epstein_files: EpsteinFiles) -> None:
    console.print(Panel('JSON Stats Dump', expand=True, style='reverse bold'), '\n')
    print_json(f"{MESSENGER_LOG_CLASS} Sender Counts", MessengerLog.count_authors(epstein_files.imessage_logs), skip_falsey=True)
    print_json(f"{EMAIL_CLASS} Author Counts", epstein_files.email_author_counts, skip_falsey=True)
    print_json(f"{EMAIL_CLASS} Recipient Counts", epstein_files.email_recipient_counts, skip_falsey=True)
    print_json("Email signature_substitution_countss", epstein_files.email_signature_substitution_counts(), skip_falsey=True)
    print_json("email_author_device_signatures", dict_sets_to_lists(epstein_files.email_authors_to_device_signatures))
    print_json("email_sent_from_devices", dict_sets_to_lists(epstein_files.email_device_signatures_to_authors))
    print_json("email_unknown_recipient_file_ids", epstein_files.email_unknown_recipient_file_ids())
    print_json("count_by_month", count_by_month(epstein_files.all_documents()))



"""
Helpers for dealing with various kinds of data.
"""
import itertools
import re
import time
from dataclasses import dataclass, field
from datetime import datetime, timezone
from dateutil import tz
from typing import TypeVar

from dateutil.parser import parse
from rich.text import Text

from epstein_files.util.constant import names
from epstein_files.util.env import args
from epstein_files.util.logging import logger

T = TypeVar('T')

ISO_DATE_REGEX = re.compile(r'\d{4}-\d{2}(-\d{2})?')
MULTINEWLINE_REGEX = re.compile(r"\n{2,}")
CONSTANT_VAR_REGEX = re.compile(r"^[A-Z_]+$")
ALL_NAMES = [v for k, v in vars(names).items() if isinstance(v, str) and CONSTANT_VAR_REGEX.match(k)]

PACIFIC_TZ = tz.gettz("America/Los_Angeles")
TIMEZONE_INFO = {"PST": PACIFIC_TZ, "PDT": PACIFIC_TZ}  # Suppresses annoying warnings from parse() calls


def dict_sets_to_lists(d: dict[str, set]) -> dict[str, list]:
    return {k: sorted(list(v)) for k, v in d.items()}


def extract_datetime(s: str) -> datetime | None:
    match = ISO_DATE_REGEX.search(s)

    if not match:
        return None

    date_str = match.group(0)

    if len(date_str) == 4:
        date_str += '-01-01'
    elif len(date_str) == 7:
        date_str += '-01'

    return parse(date_str, tzinfos=TIMEZONE_INFO)


def extract_last_name(name: str) -> str:
    if ' ' not in name:
        return name

    names = name.split()

    if names[-1].startswith('Jr') and len(names[-1]) <= 3:
        return ' '.join(names[-2:])
    else:
        return names[-1]


def flatten(_list: list[list[T]]) -> list[T]:
    return list(itertools.chain.from_iterable(_list))


def json_safe(d: dict) -> dict:
    return {
        'None' if k is None else k: v.isoformat() if isinstance(v, datetime) else v
        for k,v in d.items()
    }


def listify(listlike) -> list:
    """Create a list of 'listlike'. Returns empty list if 'listlike' is None or empty string."""
    if isinstance(listlike, list):
        return listlike
    elif listlike is None:
        return [None]
    elif listlike:
        return [listlike]
    else:
        return []


def ordinal_str(n: int) -> str:
    if 11 <= (n % 100) <= 13:
        suffix = 'th'
    else:
        suffix = {1: 'st', 2: 'nd', 3: 'rd'}.get(n % 10, 'th')

    return str(n) + suffix


def patternize(_pattern: str | re.Pattern):
    return _pattern if isinstance(_pattern, re.Pattern) else re.compile(rf"({_pattern})", re.IGNORECASE)


def remove_timezone(timestamp: datetime) -> datetime:
    if timestamp.tzinfo:
        timestamp = timestamp.astimezone(timezone.utc).replace(tzinfo=None)
        logger.debug(f"    -> Converted to UTC: {timestamp}")

    return timestamp


def sort_dict(d: dict[str | None, int] | dict[str, int]) -> list[tuple[str | None, int]]:
    sort_key = lambda e: (e[0] or '').lower() if args.sort_alphabetical else [-e[1], (e[0] or '').lower()]
    return sorted(d.items(), key=sort_key)


collapse_newlines = lambda text: MULTINEWLINE_REGEX.sub('\n\n', text)
date_str = lambda dt: dt.isoformat()[0:10] if dt else None
escape_double_quotes = lambda text: text.replace('"', r'\"')
escape_single_quotes = lambda text: text.replace("'", r"\'")
iso_timestamp = lambda dt: dt.isoformat().replace('T', ' ')
uniquify = lambda _list: list(set(_list))
without_nones = lambda _list: [e for e in _list if e]



import re
from dataclasses import dataclass, field
from datetime import datetime

from rich.text import Text

from epstein_files.util.constant.names import JEFFREY_EPSTEIN, ANTHONY_SCARAMUCCI, STEVE_BANNON, UNKNOWN
from epstein_files.util.data import extract_last_name
from epstein_files.util.highlighted_group import get_style_for_name
from epstein_files.util.logging import logger
from epstein_files.util.rich import TEXT_LINK, highlighter

MSG_DATE_FORMAT = r"%m/%d/%y %I:%M:%S %p"
PHONE_NUMBER_REGEX = re.compile(r'^[\d+]+.*')
TIMESTAMP_STYLE = 'turquoise4 dim'

DISPLAY_LAST_NAME_ONLY = [
    JEFFREY_EPSTEIN,
    STEVE_BANNON,
]

UNKNOWN_TEXTERS = [
    '+16463880059',
    '+13108737937',
    '+13108802851',
]

TEXTER_MAPPING = {
    'e:': JEFFREY_EPSTEIN,
    'e:jeeitunes@gmail.com': JEFFREY_EPSTEIN,
    '+19174393646': ANTHONY_SCARAMUCCI,
    '+13109906526': STEVE_BANNON,
}


@dataclass(kw_only=True)
class TextMessage:
    """Class representing a single iMessage text message."""
    author: str | None
    author_str: str = field(init=False)
    id_confirmed: bool = False
    text: str
    timestamp_str: str

    def __post_init__(self):
        self.author = TEXTER_MAPPING.get(self.author or UNKNOWN, self.author)

        if self.author is None:
            self.author_str = UNKNOWN
        elif self.author in UNKNOWN_TEXTERS:
            logger.warning(f"Bad text from '{self.author}': \"{self.text}\"")
            self.author_str = self.author
            self.author = None  # TODO: this shouldn't be happening; we still know the author...
        elif self.author in DISPLAY_LAST_NAME_ONLY:
            self.author_str = extract_last_name(self.author)
        else:
            self.author_str = self.author

        if not self.id_confirmed and self.author is not None and self.author != JEFFREY_EPSTEIN:
            self.author_str = self.author + ' (?)'

    def timestamp(self) -> datetime:
        return datetime.strptime(self.timestamp_str, MSG_DATE_FORMAT)

    def _message(self) -> Text:
        lines = self.text.split('\n')

        # Fix multiline links
        if self.text.startswith('http'):
            text = self.text

            if len(lines) > 1 and not lines[0].endswith('html'):
                if len(lines) > 2 and lines[1].endswith('-'):
                    text = text.replace('\n', '', 2)
                else:
                    text = text.replace('\n', '', 1)

            lines = text.split('\n')
            link_text = lines.pop()
            msg_txt = Text('').append(Text.from_markup(f"[link={link_text}]{link_text}[/link]", style=TEXT_LINK))

            if len(lines) > 0:
                msg_txt.append('\n' + ' '.join(lines))
        else:
            msg_txt = highlighter(' '.join(lines))  # remove newlines

        return msg_txt

    def __rich__(self) -> Text:
        # TODO: Workaround for phone numbers that sucks
        author_style = get_style_for_name(self.author_str if self.author_str.startswith('+') else self.author)
        author_txt = Text(self.author_str, style=author_style)
        timestamp_txt = Text(f"[{self.timestamp_str}]", style=TIMESTAMP_STYLE).append(' ')
        return Text('').append(timestamp_txt).append(author_txt).append(': ', style='dim').append(self._message())



import logging
import re
from dataclasses import asdict, dataclass, field
from datetime import datetime
from pathlib import Path
from subprocess import run
from typing import ClassVar, Sequence, TypeVar

from rich.console import Console, ConsoleOptions, Group, RenderResult
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import *
from epstein_files.util.constants import ALL_FILE_CONFIGS, FALLBACK_TIMESTAMP
from epstein_files.util.data import collapse_newlines, date_str, iso_timestamp, listify, patternize, without_nones
from epstein_files.util.doc_cfg import EmailCfg, DocCfg, Metadata, TextCfg
from epstein_files.util.env import args
from epstein_files.util.file_helper import (DOCS_DIR, file_stem_for_id, extract_file_id, file_size,
     file_size_str, is_local_extract_file)
from epstein_files.util.logging import DOC_TYPE_STYLES, FILENAME_STYLE, logger
from epstein_files.util.rich import SYMBOL_STYLE, console, highlighter, key_value_txt, link_text_obj
from epstein_files.util.search_result import MatchedLine

CLOSE_PROPERTIES_CHAR = ']'
HOUSE_OVERSIGHT = HOUSE_OVERSIGHT_PREFIX.replace('_', ' ').strip()
INFO_INDENT = 2
INFO_PADDING = (0, 0, 0, INFO_INDENT)
MAX_TOP_LINES_LEN = 4000  # Only for logging
MIN_DOCUMENT_ID = 10477
LOCAL_EXTRACT_REGEX = re.compile(r"_\d$")
WHITESPACE_REGEX = re.compile(r"\s{2,}|\t|\n", re.MULTILINE)

MIN_TIMESTAMP = datetime(1991, 1, 1)
MID_TIMESTAMP = datetime(2007, 1, 1)
MAX_TIMESTAMP = datetime(2020, 1, 1)

FILENAME_MATCH_STYLES = [
    'dark_green',
    'green',
    'spring_green4',
]

METADATA_FIELDS = [
    'author',
    'file_id',
    'num_lines',
    'timestamp'
]

OCR_REPAIRS = {
    re.compile(r'\.corn\b'): '.com',
    re.compile('ln(adequate|dyke)'): r'In\1',
    'Nil Priell': 'Nili Priell',
}


@dataclass
class Document:
    """Base class for all Epstein Files documents."""
    file_path: Path
    # Optional fields
    author: str | None = None
    config: EmailCfg | DocCfg | TextCfg | None = None
    file_id: str = field(init=False)
    filename: str = field(init=False)
    is_duplicate: bool = False
    length: int = field(init=False)
    lines: list[str] = field(init=False)
    num_lines: int = field(init=False)
    text: str = ''
    timestamp: datetime | None = None
    url_slug: str = field(init=False)  # e.g. 'HOUSE_OVERSIGHT_123456

    # Class variable overridden in JsonFile
    strip_whitespace: ClassVar[bool] = True

    def __post_init__(self):
        self.filename = self.file_path.name
        self.file_id = extract_file_id(self.filename)
        self.config = ALL_FILE_CONFIGS.get(self.file_id)
        self.is_duplicate = bool(self.config.dupe_of_id) if self.config else False

        if self.is_local_extract_file():
            self.url_slug = LOCAL_EXTRACT_REGEX.sub('', file_stem_for_id(self.file_id))
            cfg_type = type(self.config).__name__ if self.config else None

            # Coerce FileConfig for court docs etc. to MessageCfg for email files extracted from that document
            if self.class_name() == EMAIL_CLASS and self.config and cfg_type != EmailCfg.__name__:
                self.config = EmailCfg.from_doc_cfg(self.config)
        else:
            self.url_slug = self.file_path.stem

        self._set_computed_fields(text=self.text or self._load_file())
        self._repair()
        self._extract_author()
        self.timestamp = self._extract_timestamp()

    def class_name(self) -> str:
        """Annoying workaround for circular import issues and isinstance()."""
        return str(type(self).__name__)

    def configured_description(self) -> str | None:
        """Overloaded in OtherFile."""
        if self.config and self.config.description:
            return f"({self.config.description})"

    def date_str(self) -> str | None:
        return date_str(self.timestamp)

    def description_panel(self, include_hints: bool = False) -> Panel:
        """Panelized description() with info_txt(), used in search results."""
        hints = [Text('', style='italic').append(h) for h in (self.hints() if include_hints else [])]
        return Panel(Group(*([self.summary()] + hints)), border_style=self.document_type_style(), expand=False)

    def document_type_style(self) -> str:
        return DOC_TYPE_STYLES[self.class_name()]

    def duplicate_file_txt(self) -> Text:
        """If the file is a dupe make a nice message to explain what file it's a duplicate of."""
        if not self.config or not self.config.dupe_of_id:
            raise RuntimeError(f"duplicate_file_txt() called on {self.summary()} but not a dupe! config:\n\n{self.config}")

        txt = Text(f"Not showing ", style='white dim italic').append(epstein_media_doc_link_txt(self.file_id, style='cyan'))
        txt.append(f" because it's {self.config.duplicate_reason()} ")
        return txt.append(epstein_media_doc_link_txt(self.config.dupe_of_id, style='royal_blue1'))

    def epsteinify_link(self, style: str = ARCHIVE_LINK_COLOR, link_txt: str | None = None) -> Text:
        """Create a Text obj link to this document on epsteinify.com."""
        return link_text_obj(epsteinify_doc_url(self.url_slug), link_txt or self.url_slug, style)

    def epstein_media_link(self, style: str = ARCHIVE_LINK_COLOR, link_txt: str | None = None) -> Text:
        """Create a Text obj link to this document on epstein.media."""
        return link_text_obj(epstein_media_doc_url(self.url_slug), link_txt or self.url_slug, style)

    def epstein_web_link(self, style: str = ARCHIVE_LINK_COLOR, link_txt: str | None = None) -> Text:
        """Create a Text obj link to this document on EpsteinWeb."""
        return link_text_obj(epstein_web_doc_url(self.url_slug), link_txt or self.url_slug, style)

    def file_info_panel(self) -> Group:
        """Panel with filename linking to raw file plus any hints/info about the file."""
        panel = Panel(self.raw_document_link_txt(include_alt_link=True), border_style=self._border_style(), expand=False)
        hints = [Padding(hint, INFO_PADDING) for hint in self.hints()]
        return Group(*([panel] + hints))

    def file_size(self) -> int:
        return file_size(self.file_path)

    def file_size_str(self) -> str:
        return file_size_str(self.file_path)

    def hints(self) -> list[Text]:
        """Additional info about the Document (author, description, and so on) to be desplayed in doc header."""
        hints = listify(self.info_txt())
        hint_msg = self.configured_description()

        if hint_msg:
            hints.append(highlighter(Text(hint_msg, style='white dim italic')))

        return without_nones(hints)

    def info_txt(self) -> Text | None:
        """Secondary info about this file (recipients, level of certainty, etc). Overload in subclasses."""
        return None

    def is_local_extract_file(self) -> bool:
        """True if file created by extracting text from a court doc (identifiable from filename e.g. HOUSE_OVERSIGHT_012345_1.txt)."""
        return is_local_extract_file(self.filename)

    def log(self, msg: str, level: int = logging.WARNING):
        """Log with filename as a prefix."""
        logger.log(level, f"{self.url_slug} {msg}")

    def log_top_lines(self, n: int = 10, msg: str = '', level: int = logging.INFO) -> None:
        """Log first 'n' lines of self.text at 'level'. 'msg' can be optionally provided."""
        separator = '\n\n' if '\n' in msg else '. '
        msg = (msg + separator) if msg else ''
        msg = f"{self.filename}: {msg}First {n} lines:"
        logger.log(level, f"{msg}\n\n{self.top_lines(n)}\n")

    def matching_lines(self, _pattern: re.Pattern | str) -> list[MatchedLine]:
        """Return lines matching a regex as colored list[Text]."""
        pattern = patternize(_pattern)
        return [MatchedLine(line, i) for i, line in enumerate(self.lines) if pattern.search(line)]

    def metadata(self) -> Metadata:
        metadata = self.config.metadata() if self.config else {}
        metadata.update({k: v for k, v in asdict(self).items() if k in METADATA_FIELDS and v is not None})
        metadata['bytes'] = self.file_size()
        metadata['filename'] = f"{self.url_slug}.txt"
        metadata['type'] = self.class_name()

        if self.is_local_extract_file():
            metadata['extracted_file'] = {
                'explanation': 'This file was extracted from a court filing, not distributed directly. A copy can be found on github.',
                'extracted_from_file': self.url_slug + '.txt',
                'extracted_file_url': extracted_file_url(self.filename),
            }

        return metadata

    def raw_text(self) -> str:
        with open(self.file_path) as f:
            return f.read()

    def raw_document_link_txt(self, style: str = '', include_alt_link: bool = False) -> Text:
        """Returns colored links to epstein.media and and epsteinweb in a Text object."""
        txt = Text('', style='white' if include_alt_link else ARCHIVE_LINK_COLOR)

        if args.use_epstein_web_links:
            txt.append(self.epstein_web_link(style=style))

            if include_alt_link:
                txt.append(' (').append(self.epsteinify_link(style='white dim', link_txt=EPSTEINIFY)).append(')')
                txt.append(' (').append(self.epstein_media_link(style='white dim', link_txt=EPSTEIN_MEDIA)).append(')')
        else:
            txt.append(self.epstein_media_link(style=style))

            if include_alt_link:
                txt.append(' (').append(self.epsteinify_link(style='white dim', link_txt=EPSTEINIFY)).append(')')
                txt.append(' (').append(self.epstein_web_link(style='white dim', link_txt=EPSTEIN_WEB)).append(')')

        return txt

    def repair_ocr_text(self, repairs: dict[str | re.Pattern, str], text: str) -> str:
        """Apply a dict of repairs (key is pattern or string, value is replacement string) to text."""
        for k, v in repairs.items():
            if isinstance(k, re.Pattern):
                text = k.sub(v, text)
            else:
                text = text.replace(k, v)

        return text

    def sort_key(self) -> tuple[datetime, str, int]:
        if self.config and self.config.dupe_of_id:
            sort_id = self.config.dupe_of_id
            dupe_idx = 1
        else:
            sort_id = self.file_id
            dupe_idx = 0

        return (self.timestamp or FALLBACK_TIMESTAMP, sort_id, dupe_idx)

    def summary(self) -> Text:
        """Summary of this file for logging. Brackets are left open for subclasses to add stuff."""
        txt = Text('').append(self.class_name(), style=self.document_type_style())
        txt.append(f" {self.url_slug}", style=FILENAME_STYLE)

        if self.timestamp:
            timestamp_str = iso_timestamp(self.timestamp).removesuffix(' 00:00:00')
            txt.append(' (', style=SYMBOL_STYLE)
            txt.append(f"{timestamp_str}", style=TIMESTAMP_DIM).append(')', style=SYMBOL_STYLE)

        txt.append(' [').append(key_value_txt('size', Text(self.file_size_str(), style='aquamarine1')))
        txt.append(", ").append(key_value_txt('lines', Text(f"{self.num_lines}", style='cyan')))
        return txt

    def top_lines(self, n: int = 10) -> str:
        return '\n'.join(self.lines[0:n])[:MAX_TOP_LINES_LEN]

    def _border_style(self) -> str:
        """Should be overloaded in subclasses."""
        return 'white'

    def _extract_author(self) -> None:
        """Get author from config. Extended in Email subclass to also check headers."""
        if self.config and self.config.author:
            self.author = self.config.author

    def _extract_timestamp(self) -> datetime | None:
        """Should be implemented in subclasses."""
        pass

    def _load_file(self) -> str:
        """Remove BOM and HOUSE OVERSIGHT lines, strip whitespace."""
        text = self.raw_text()
        text = text[1:] if (len(text) > 0 and text[0] == '\ufeff') else text  # remove BOM
        text = self.repair_ocr_text(OCR_REPAIRS, text.strip())
        lines = [l.strip() for l in text.split('\n') if not l.startswith(HOUSE_OVERSIGHT)]
        return collapse_newlines('\n'.join(lines))

    def _repair(self) -> None:
        """Can optionally be overloaded in subclasses to further improve self.text."""
        pass

    def _set_computed_fields(self, lines: list[str] | None = None, text: str | None = None) -> None:
        """Sets all fields derived from self.text based on either 'lines' or 'text' arg."""
        if (lines and text):
            raise RuntimeError(f"[{self.filename}] Either 'lines' or 'text' arg must be provided (got both)")
        elif lines is not None:
            self.text = '\n'.join(lines).strip()
        elif text is not None:
            self.text = text.strip()
        else:
            raise RuntimeError(f"[{self.filename}] Either 'lines' or 'text' arg must be provided (neither was)")

        self.length = len(self.text)
        self.lines = [line.strip() if self.strip_whitespace else line for line in self.text.split('\n')]
        self.num_lines = len(self.lines)

    def _write_clean_text(self, output_path: Path) -> None:
        """Write self.text to 'output_path'. Used only for diffing files."""
        if output_path.exists():
            if str(output_path.name).startswith(HOUSE_OVERSIGHT_PREFIX):
                raise RuntimeError(f"'{output_path}' already exists! Not overwriting.")
            else:
                logger.warning(f"Overwriting '{output_path}'...")

        with open(output_path, 'w') as f:
            f.write(self.text)

        logger.warning(f"Wrote {self.length} chars of cleaned {self.filename} to {output_path}.")

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield self.file_info_panel()
        text_panel = Panel(highlighter(self.text), border_style=self._border_style(), expand=False)
        yield Padding(text_panel, (0, 0, 1, INFO_INDENT))

    def __str__(self) -> str:
        return self.summary().plain

    @staticmethod
    def diff_files(files: list[str]) -> None:
        """Diff the contents of two Documents after all cleanup, BOM removal, etc."""
        if len(files) != 2:
            raise RuntimeError('Need 2 files')
        elif files[0] == files[1]:
            raise RuntimeError(f"Filenames are the same!")

        files = [f"{HOUSE_OVERSIGHT_PREFIX}{f}" if len(f) == 6 else f for f in files]
        files = [f if f.endswith('.txt') else f"{f}.txt" for f in files]
        tmpfiles = [Path(f"tmp_{f}") for f in files]
        docs = [Document(DOCS_DIR.joinpath(f)) for f in files]

        for i, doc in enumerate(docs):
            doc._write_clean_text(tmpfiles[i])

        cmd = f"diff {tmpfiles[0]} {tmpfiles[1]}"
        console.print(f"Running '{cmd}'...")
        results = run(cmd, shell=True, capture_output=True, text=True).stdout

        for line in _color_diff_output(results):
            console.print(line, highlight=True)

        console.print(f"Possible suppression with: ")
        console.print(Text('   suppress left: ').append(f"   '{extract_file_id(files[0])}': 'the same as {extract_file_id(files[1])}',", style='cyan'))
        console.print(Text('  suppress right: ').append(f"   '{extract_file_id(files[1])}': 'the same as {extract_file_id(files[0])}',", style='cyan'))

        for f in tmpfiles:
            f.unlink()

    @staticmethod
    def sort_by_timestamp(docs: Sequence['DocumentType']) -> list['DocumentType']:
        return sorted(docs, key=lambda doc: doc.sort_key())

    @classmethod
    def uniquify(cls, documents: Sequence['DocumentType']) -> Sequence['DocumentType']:
        """Uniquify by file_id."""
        id_map = {doc.file_id: doc for doc in documents}
        return [doc for doc in id_map.values()]


DocumentType = TypeVar('DocumentType', bound=Document)


def _color_diff_output(diff_result: str) -> list[Text]:
    txts = [Text('diff output:')]
    style = 'dim'

    for line in diff_result.split('\n'):
        if line.startswith('>'):
            style='spring_green4'
        elif line.startswith('<'):
            style='sea_green1'

        txts.append(Text(line, style=style))

    return txts



import re
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime

from rich.console import Console, ConsoleOptions, RenderResult
from rich.table import Table
from rich.text import Text

from epstein_files.documents.communication import Communication
from epstein_files.documents.imessage.text_message import MSG_DATE_FORMAT, TextMessage
from epstein_files.util.constant.names import JEFFREY_EPSTEIN, UNKNOWN
from epstein_files.util.constant.strings import AUTHOR
from epstein_files.util.data import iso_timestamp, listify, sort_dict
from epstein_files.util.doc_cfg import Metadata, TextCfg
from epstein_files.util.highlighted_group import get_style_for_name
from epstein_files.util.logging import logger

CONFIRMED_MSG = 'Found confirmed counterparty'
GUESSED_MSG = 'This is probably a conversation with'
MSG_REGEX = re.compile(r'Sender:(.*?)\nTime:(.*? (AM|PM)).*?Message:(.*?)\s*?((?=(\nSender)|\Z))', re.DOTALL)
REDACTED_AUTHOR_REGEX = re.compile(r"^([-+•_1MENO.=F]+|[4Ide])$")


@dataclass
class MessengerLog(Communication):
    """Class representing one iMessage log file (one conversation between Epstein and some counterparty)."""
    config: TextCfg | None = None
    _messages: list[TextMessage] = field(default_factory=list)

    def first_message_at(self, name: str | None) -> datetime:
        return self.messages_by(name)[0].timestamp()

    def info_txt(self) -> Text | None:
        if self.author is None:
            return None

        hint_msg = GUESSED_MSG if self.is_attribution_uncertain() else CONFIRMED_MSG
        author_txt = Text(self.author_or_unknown(), style=self.author_style + ' bold')
        return Text(f"({hint_msg} ", style='dim').append(author_txt).append(')')

    def last_message_at(self, name: str | None) -> datetime:
        return self.messages_by(name)[-1].timestamp()

    def messages(self) -> list[TextMessage]:
        """Lazily evaluated accessor for self._messages."""
        if len(self._messages) == 0:
            self._messages = [
                TextMessage(
                    # If the Sender: is redacted that means it's from self.author
                    author=REDACTED_AUTHOR_REGEX.sub('', match.group(1).strip()) or self.author,
                    id_confirmed=not self.is_attribution_uncertain(),
                    text=match.group(4).strip(),
                    timestamp_str=match.group(2).strip(),
                )
                for match in MSG_REGEX.finditer(self.text)
            ]

        return self._messages

    def messages_by(self, name: str | None) -> list[TextMessage]:
        """Return all messages by 'name'."""
        return [m for m in self.messages() if m.author == name]

    def metadata(self) -> Metadata:
        metadata = super().metadata()
        metadata.update({'num_messages': len(self.messages())})
        return metadata

    def _border_style(self) -> str:
        return self.author_style

    def _extract_timestamp(self) -> datetime:
        for match in MSG_REGEX.finditer(self.text):
            timestamp_str = match.group(2).strip()

            try:
                return datetime.strptime(timestamp_str, MSG_DATE_FORMAT)
            except ValueError as e:
                logger.info(f"Failed to parse '{timestamp_str}' to datetime! Using next match. Error: {e}'")

        raise RuntimeError(f"{self}: No timestamp found!")

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield self.file_info_panel()
        yield Text('')

        for message in self.messages():
            yield message

    @classmethod
    def count_authors(cls, imessage_logs: list['MessengerLog']) -> dict[str | None, int]:
        """Count up how many texts were sent by each author."""
        sender_counts: dict[str | None, int] = defaultdict(int)

        for message_log in imessage_logs:
            for message in message_log.messages():
                sender_counts[message.author] += 1

        return sender_counts

    @classmethod
    def logs_for(cls, author: str | None | list[str | None], logs: list['MessengerLog']) -> list['MessengerLog']:
        authors = listify(author)
        return logs if JEFFREY_EPSTEIN in authors else [log for log in logs if log.author in authors]

    @classmethod
    def summary_table(cls, imessage_logs: list['MessengerLog']) -> Table:
        """Build a table summarizing the text messages in 'imessage_logs'."""
        counts_table = Table(title="Text Message Counts By Author", header_style="bold")
        counts_table.add_column(AUTHOR.title(), justify='left', style="steel_blue bold", width=30)
        counts_table.add_column('Files', justify='right', style='white')
        counts_table.add_column("Msgs", justify='right')
        counts_table.add_column('First Sent At', justify='center', highlight=True, width=21)
        counts_table.add_column('Last Sent At', justify='center', style='wheat4', width=21)
        counts_table.add_column('Days', justify='right', style='dim')

        for name, count in sort_dict(cls.count_authors(imessage_logs)):
            logs = cls.logs_for(name, imessage_logs)
            first_at = logs[0].first_message_at(name)
            last_at = logs[-1].first_message_at(name)

            counts_table.add_row(
                Text(name or UNKNOWN, get_style_for_name(name)),
                str(len(logs)),
                f"{count:,}",
                iso_timestamp(first_at),
                iso_timestamp(last_at),
                str((last_at - first_at).days + 1),
            )

        return counts_table



import re
import logging
import warnings
from dataclasses import dataclass
from datetime import datetime

import datefinder
import dateutil
from rich.console import Group
from rich.markup import escape
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.document import CLOSE_PROPERTIES_CHAR, WHITESPACE_REGEX, Document
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import *
from epstein_files.util.doc_cfg import FINANCIAL_REPORTS_AUTHORS, DocCfg
from epstein_files.util.data import escape_single_quotes, remove_timezone, uniquify
from epstein_files.util.file_helper import FILENAME_LENGTH
from epstein_files.util.env import args
from epstein_files.util.highlighted_group import get_style_for_category
from epstein_files.util.rich import QUESTION_MARK_TXT, highlighter
from epstein_files.util.logging import logger

MAX_DAYS_SPANNED_TO_BE_VALID = 10
MAX_EXTRACTED_TIMESTAMPS = 100
MIN_TIMESTAMP = datetime(2000, 1, 1)
MID_TIMESTAMP = datetime(2007, 1, 1)
MAX_TIMESTAMP = datetime(2022, 12, 31)
PREVIEW_CHARS = int(580 * (1 if args.all_other_files else 1.5))
LOG_INDENT = '\n         '
TIMESTAMP_LOG_INDENT = f'{LOG_INDENT}    '
VAST_HOUSE = 'vast house'  # Michael Wolff article draft about Epstein indicator
VI_DAILY_NEWS_REGEX = re.compile(r'virgin\s*is[kl][ai]nds\s*daily\s*news', re.IGNORECASE)

UNINTERESTING_CATEGORES = [
    ARTS,
    BOOK,
    JUNK,
    SPEECH,
]

UNINTERESTING_IDS = [
    '031794',
]

# OtherFiles whose description/hints match these prefixes are not displayed unless --all-other-files is used
UNINTERESTING_PREFIXES = FINANCIAL_REPORTS_AUTHORS + [
    'article about',
    ARTICLE_DRAFT,
    'Aviation International',
    BBC,
    BLOOMBERG,
    'Boston Globe',
    BROCKMAN_INC,
    CHINA_DAILY,
    CNN,
    'completely redacted',
    CVRA,
    DAILY_MAIL,
    DAILY_TELEGRAPH,
    DAVID_SCHOEN_CVRA_LEXIS_SEARCH[0:-12],  # Because date at end :(
    DERSH_GIUFFRE_TWEET,
    'Financial Times',
    'Forbes',
    'Frontlines',
    'Future Science',
    'Globe and Mail',
    GORDON_GETTY,
    f"{HARVARD} Econ",
    HARVARD_POETRY,
    'Inference',
    JASTA,
    'JetGala',
    JOHN_BOLTON_PRESS_CLIPPING,
    'Journal of Criminal',
    LA_TIMES,
    'Litigation Daily',
    LAWRENCE_KRAUSS,
    'MarketWatch',
    MARTIN_NOWAK,
    NOBEL_CHARITABLE_TRUST,
    'Nautilus',
    'New Yorker',
    NYT_ARTICLE,
    NYT_COLUMN,
    PALM_BEACH_CODE_ENFORCEMENT,
    PALM_BEACH_DAILY_ARTICLE,
    PALM_BEACH_POST_ARTICLE,
    PALM_BEACH_TSV,
    PALM_BEACH_WATER_COMMITTEE,
    PAUL_KRASSNER,
    PEGGY_SIEGAL,
    'Politifact',
    'Rafanelli',
    ROBERT_LAWRENCE_KUHN,
    ROBERT_TRIVERS,
    'SCMP',
    'SciencExpress',
    'Scowcroft',
    SHIMON_POST_ARTICLE,
    SINGLE_PAGE,
    STACEY_PLASKETT,
    TERJE_ROD_LARSEN,
    TEXT_OF_US_LAW,
    TRANSLATION,
    TWEET,
    THE_REAL_DEAL_ARTICLE,
    TRUMP_DISCLOSURES,
    UBS_CIO_REPORT,
    UN_GENERAL_ASSEMBLY,
    'U.S. News',
    'US Office',
    'Vanity Fair',
    VI_DAILY_NEWS_ARTICLE,
    WAPO,
]


@dataclass
class OtherFile(Document):
    """File that is not an email, an iMessage log, or JSON data."""

    def __post_init__(self):
        super().__post_init__()

        if self.config is None and VI_DAILY_NEWS_REGEX.search(self.text):
            self.log(f"Creating synthetic config for VI Daily News article...", logging.INFO)
            self.config = DocCfg(id=self.file_id, description=VI_DAILY_NEWS_ARTICLE, category=ARTICLE)

    def category(self) -> str | None:
        return self.config and self.config.category

    def configured_description(self) -> str | None:
        """Overloads superclass method."""
        if self.config is not None:
            return self.config.info_str()

    def description_panel(self, include_hints=True) -> Panel:
        """Panelized description() with info_txt(), used in search results."""
        return super().description_panel(include_hints=include_hints)

    def highlighted_preview_text(self) -> Text:
        try:
            return highlighter(escape(self.preview_text()))
        except Exception as e:
            logger.error(f"Failed to apply markup in string '{escape_single_quotes(self.preview_text())}'\n"
                         f"Original string: '{escape_single_quotes(self.preview_text())}'\n"
                         f"File: '{self.filename}'\n")

            return Text(escape(self.preview_text()))

    def is_interesting(self):
        """False for lame prefixes, duplicates, and other boring files."""
        hints = self.hints()

        if self.is_duplicate:
            return False
        elif self.file_id in UNINTERESTING_IDS:
            return False
        elif len(hints) == 0:
            return True
        elif self.config:
            if self.config.is_interesting:
                return True
            elif self.category() == FINANCE and self.author is not None:
                return False
            elif self.category() in UNINTERESTING_CATEGORES:
                return False

        for prefix in UNINTERESTING_PREFIXES:
            if hints[0].plain.startswith(prefix):
                return False

        return True

    def preview_text(self) -> str:
        return WHITESPACE_REGEX.sub(' ', self.text)[0:PREVIEW_CHARS]

    def summary(self) -> Text:
        """One line summary mostly for logging."""
        return super().summary().append(CLOSE_PROPERTIES_CHAR)

    def _extract_timestamp(self) -> datetime | None:
        """Return configured timestamp or value extracted by scanning text with datefinder."""
        if self.config and self.config.timestamp:
            return self.config.timestamp

        timestamps: list[datetime] = []

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="datefinder")
            warnings.filterwarnings("ignore", module="dateutil")

            try:
                for timestamp in datefinder.find_dates(self.text, strict=True):
                    timestamp = remove_timezone(timestamp)

                    if MIN_TIMESTAMP < timestamp < MAX_TIMESTAMP:
                        timestamps.append(timestamp)

                    if len(timestamps) >= MAX_EXTRACTED_TIMESTAMPS:
                        break
            except ValueError as e:
                logger.warning(f"Error while iterating through datefinder.find_dates(): {e}")

        if len(timestamps) == 0:
            if not self.is_duplicate and VAST_HOUSE not in self.text:
                self.log_top_lines(15, msg=f"No timestamps found", level=logging.INFO)

            return None
        elif len(timestamps) == 1:
            return timestamps[0]
        else:
            timestamps = sorted(uniquify(timestamps), reverse=True)
            self._log_extracted_timestamps_info(timestamps)
            return timestamps[0]  # Most recent timestamp appearing in text is usually the closest

    def _log_extracted_timestamps_info(self, timestamps: list[datetime]) -> None:
        num_days_spanned = (timestamps[0] - timestamps[-1]).days
        timestamps_log_msg = f"Extracted {len(timestamps)} timestamps spanning {num_days_spanned} days{TIMESTAMP_LOG_INDENT}"
        timestamps_log_msg += TIMESTAMP_LOG_INDENT.join([str(dt) for dt in timestamps])

        if num_days_spanned > MAX_DAYS_SPANNED_TO_BE_VALID and VAST_HOUSE not in self.text:
            self.log_top_lines(15, msg=timestamps_log_msg, level=logging.DEBUG)

    @staticmethod
    def build_table(docs: list['OtherFile']) -> Table:
        """Build a table of OtherFile documents."""
        table = Table(header_style='bold', show_lines=True)
        table.add_column('File', justify='center', width=FILENAME_LENGTH)
        table.add_column('Date', justify='center')
        table.add_column('Size', justify='center')
        table.add_column('Type', justify='center')
        table.add_column(FIRST_FEW_LINES, justify='left', style='pale_turquoise4')

        for doc in docs:
            link_and_info = [doc.raw_document_link_txt()]
            category = doc.category()
            date_str = doc.date_str()

            if doc.is_duplicate:
                preview_text = doc.duplicate_file_txt()
                row_style = ' dim'
            else:
                link_and_info += doc.hints()
                preview_text = doc.highlighted_preview_text()
                row_style = ''

            if category:
                category_txt = Text(category, get_style_for_category(category) or 'wheat4')
            else:
                category_txt = Text('')

            table.add_row(
                Group(*link_and_info),
                Text(date_str, style=TIMESTAMP_DIM) if date_str else QUESTION_MARK_TXT,
                doc.file_size_str(),
                category_txt,
                preview_text,
                style=row_style
            )

        return table



import re
from copy import deepcopy
from dataclasses import Field, asdict, dataclass, field, fields
from datetime import datetime
from typing import Generator, Literal

from dateutil.parser import parse

from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.data import without_nones

DuplicateType = Literal['earlier', 'quoted', 'redacted', 'same']
Metadata = dict[str, bool | datetime | int | str | list[str | None] |dict[str, bool | str]]

# Misc
CONSTANTIZE_NAMES = False  # A flag set to True that causes repr() of these classes to return strings of usable code
INDENT = '    '
INDENT_NEWLINE = f'\n{INDENT}'
INDENTED_JOIN = f',{INDENT_NEWLINE}'
MAX_LINE_LENGTH = 150
REPUTATION_MGMT = f'{REPUTATION} management'
SAME = 'same'

DUPE_TYPE_STRS: dict[DuplicateType, str] = {
    'earlier': 'an earlier draft of',
    'quoted': 'quoted in full in',
    'redacted': 'a redacted version of',
    SAME: 'the same as',
}

FIELD_SORT_KEY = {
    'id': 'a',
    'author': 'aa',
    'attribution_reason': 'zz',
}

FINANCIAL_REPORTS_AUTHORS = [
    BOFA,
    DEUTSCHE_BANK,
    ELECTRON_CAPITAL_PARTNERS,
    GOLDMAN_INVESTMENT_MGMT,
    'Invesco',
    JP_MORGAN,
    'Morgan Stanley',
    'S&P',
]

# Fields like timestamp and author are better added from the Document object
INVALID_FOR_METADATA = [
    'actual_text',
    'date',
    'id',
    'timestamp',
    'was_generated',
]


@dataclass(kw_only=True)
class DocCfg:
    """
    Encapsulates info about files that needs to be manually configured because it cannot be programmatically inferred.

    Attributes:
        id (str): ID of file
        author (str | None): Author of the document (if any)
        category (str | None): Type of file
        date (str | None): If passed will be immediated parsed into the 'timestamp' field
        dupe_of_id (str | None): If this is a dupe the ID of the duplicated file. This file will be suppressed
        dupe_type (DuplicateType | None): The type of duplicate this file is or its 'duplicate_ids' are
        duplicate_ids (list[str]): Inverse of 'dupe_of_id' - this file will NOT be suppressed but 'duplicate_ids' will be
        is_interesting (bool): Override other considerations and always consider this file interesting
        timestamp (datetime | None): Time this email was sent, file was created, article published, etc.
        was_generated (bool): True if this object was generated by the duplicate_cfgs() method
    """
    id: str
    author: str | None = None
    category: str | None = None
    date: str | None = None
    description: str | None = None
    dupe_of_id: str | None = None
    dupe_type: DuplicateType | None = None
    duplicate_ids: list[str] = field(default_factory=list)
    is_interesting: bool = False
    timestamp: datetime | None = None
    was_generated: bool = False

    def __post_init__(self):
        if self.date:
            self.timestamp = parse(self.date)

        if self.dupe_of_id or self.duplicate_ids:
            self.dupe_type = self.dupe_type or SAME

    def duplicate_reason(self) -> str | None:
        if self.dupe_type is not None:
            return DUPE_TYPE_STRS[self.dupe_type]

    def duplicate_cfgs(self) -> Generator['DocCfg', None, None]:
        """Create synthetic DocCfg objects that set the 'dupe_of_id' field to point back to this object."""
        for id in self.duplicate_ids:
            dupe_cfg = deepcopy(self)
            dupe_cfg.id = id
            dupe_cfg.dupe_of_id = self.id
            dupe_cfg.duplicate_ids = []
            dupe_cfg.dupe_type = self.dupe_type
            dupe_cfg.was_generated = True
            yield dupe_cfg

    def info_str(self) -> str | None:
        """String that summarizes what is known about this document."""
        if self.category == REPUTATION:
            return f"{REPUTATION_MGMT}: {self.description}"
        elif self.author and self.description:
            if self.category in [ACADEMIA, BOOK]:
                return self.title_by_author()
            elif self.category == FINANCE and self.author in FINANCIAL_REPORTS_AUTHORS:
                return f"{self.author} report: '{self.description}'"
        elif self.category and self.author is None and self.description is None:
            return self.category

        pieces = without_nones([self.author, self.description])
        return ' '.join(pieces) if pieces else None

    def metadata(self) -> Metadata:
        non_null_fields = {k: v for k, v in asdict(self).items() if v and k not in INVALID_FOR_METADATA}

        if self.category in [EMAIL, TEXT_MESSAGE]:
            del non_null_fields['category']

        return non_null_fields

    def non_null_field_names(self) -> list[str]:
        return [f.name for f in self.sorted_fields() if getattr(self, f.name)]

    def sorted_fields(self) -> list[Field]:
        return sorted(fields(self), key=lambda f: FIELD_SORT_KEY.get(f.name, f.name))

    def title_by_author(self) -> str:
        if not (self.author and self.description):
            raise RuntimeError(f"Can't call title_by_author() without author and description!")

        title = self.description if '"' in self.description else f"'{self.description}'"
        return f"{title} by {self.author}"

    def _props_strs(self) -> list[str]:
        props = []
        add_prop = lambda f, value: props.append(f"{f.name}={value}")

        for _field in self.sorted_fields():
            value = getattr(self, _field.name)

            if value is None or value is False or (isinstance(value, list) and len(value) == 0):
                continue
            elif _field.name == AUTHOR:
                add_prop(_field, constantize_name(str(value)) if CONSTANTIZE_NAMES else f"'{value}'")
            elif _field.name == 'category' and value in [EMAIL, TEXT_MESSAGE]:
                continue
            elif _field.name == 'recipients' and isinstance(value, list):
                recipients_str = str([constantize_name(r) if (CONSTANTIZE_NAMES and r) else r for r in value])
                add_prop(_field, recipients_str.replace("'", '') if CONSTANTIZE_NAMES else recipients_str)
            elif _field.name == 'timestamp' and self.date is not None:
                continue  # Don't print both timestamp and date
            elif isinstance(value, datetime):
                value_str = re.sub(' 00:00:00', '', str(value))
                add_prop(_field, f"parse('{value_str}')" if CONSTANTIZE_NAMES else f"'{value}'")
            elif isinstance(value, str):
                if "'" in value:
                    value = '"' + value.replace('"', r'\"') + '"'
                else:
                    value = "'" + value.replace("'", r'\'') + "'"

                add_prop(_field, value)
            else:
                add_prop(_field, str(value))

        return props

    def __eq__(self, other: 'DocCfg') -> bool:
        """Return True if everything matches other than the two 'dupe_' fields ('duplicate_ids' is compared)."""
        for _field in self.sorted_fields():
            if _field.name == 'id' or _field.name.startswith('dupe'):
                continue
            elif getattr(self, _field.name) != getattr(other, _field.name):
                return False

        return True

    def __repr__(self) -> str:
        props = self._props_strs()
        type_str = f"{type(self).__name__}("
        single_line_repr = type_str + ', '.join(props) + f')'

        if len(single_line_repr) < MAX_LINE_LENGTH:
            repr_str = single_line_repr
        else:
            repr_str = f"{type_str}{INDENT_NEWLINE}" + INDENTED_JOIN.join(props)
            repr_str += ',' if props else ''
            repr_str += '\n)'

        if CONSTANTIZE_NAMES:
            repr_str = INDENT + INDENT_NEWLINE.join(repr_str.split('\n'))
            return repr_str.replace(',,', ',').replace(',),', '),').replace(',),', '),')
        else:
            return repr_str


@dataclass(kw_only=True)
class CommunicationCfg(DocCfg):
    """
    Manual config is always required for MessengerLog author attribution. It's also often needed for Email
    files to handle the terrible OCR text that Congress provided which messes up a lot of the email headers.

    Attributes:
        attribution_reason (str | None): Optional explanation of why this email was attributed to this author.
        is_attribution_uncertain (bool): True if we have a good idea of who the author is but are not 100% certain
    """
    attribution_reason: str | None = None
    is_attribution_uncertain: bool = False

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(kw_only=True)
class EmailCfg(CommunicationCfg):
    """
    Attributes:
        actual_text (str | None): In dire cases of broken OCR we just configure the body of the email as a string.
        is_fwded_article (bool): True if this is a newspaper article someone fwded. Used to exclude articles from word counting.
        recipients (list[str | None]): Who received the email
    """
    actual_text: str | None = None  # Override for the Email._actual_text() method for particularly broken emails
    is_fwded_article: bool = False
    recipients: list[str | None] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.category = EMAIL

    @classmethod
    def from_doc_cfg(cls, cfg: DocCfg) -> 'EmailCfg':
        return cls(**asdict(cfg))

    # This is necessary for some dumb reason. @dataclass(repr=False) doesn't cut it
    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(kw_only=True)
class TextCfg(CommunicationCfg):
    def __post_init__(self):
        super().__post_init__()
        self.category = TEXT_MESSAGE

    # This is necessary for some dumb reason. @dataclass(repr=False) doesn't cut it
    def __repr__(self) -> str:
        return super().__repr__()



import logging
from argparse import ArgumentParser
from os import environ
from pathlib import Path
from sys import argv

from epstein_files.util.logging import datefinder_logger, env_log_level, logger

DEFAULT_WIDTH = 154
HTML_SCRIPTS = ['generate_html.py', 'count_words.py']


parser = ArgumentParser(description="Parse epstein OCR docs and generate HTML page.")
parser.add_argument('--build', '-b', action='store_true', help='write HTML to docs/index.html')
parser.add_argument('--all-emails', '-ae', action='store_true', help='all the emails instead of just the interesting ones')
parser.add_argument('--all-other-files', '-ao', action='store_true', help='all the non-email, non-text msg files instead of just interesting ones')
parser.add_argument('--colors-only', '-c', action='store_true', help='print header with color key table and links and exit')
parser.add_argument('--name', '-n', action='append', dest='names', help='specify the name(s) whose communications should be output')
parser.add_argument('--output-emails', '-oe', action='store_true', help='generate other files section')
parser.add_argument('--output-other-files', '-oo', action='store_true', help='generate other files section')
parser.add_argument('--output-texts', '-ot', action='store_true', help='generate other files section')
parser.add_argument('--pickled', '-p', action='store_true', help='use pickled EpsteinFiles object')
parser.add_argument('--overwrite-pickle', '-op', action='store_true', help='generate new pickled EpsteinFiles object')
parser.add_argument('--raw', '-r', action='store_true', help='show raw contents of file (only used by scripts)')
parser.add_argument('--sort-alphabetical', '-alpha', action='store_true', help='sort emailers alphabetically in counts table')
parser.add_argument('--suppress-output', '-s', action='store_true', help='no output to terminal (use with --build)')
parser.add_argument('--use-epstein-web-links', '-use', action='store_true', help='use epsteinweb.org links instead of epstein.media')
parser.add_argument('--width', '-w', type=int, default=DEFAULT_WIDTH, help='screen width to use')
parser.add_argument('--whole-file', '-wf', action='store_true', help='print whole file (only used by search script)')
parser.add_argument('--debug', '-d', action='store_true', help='set debug level to INFO')
parser.add_argument('--deep-debug', '-dd', action='store_true', help='set debug level to DEBUG')
parser.add_argument('--make-clean', '-mc', action='store_true', help='delete all build artifact HTML and JSON files')
parser.add_argument('--suppress-logs', '-sl', action='store_true', help='set debug level to FATAL')
parser.add_argument('--json-metadata', '-jm', action='store_true', help='dump JSON metadata for all files')
parser.add_argument('--json-stats', '-j', action='store_true', help='print JSON formatted stats at the end')
parser.add_argument('positional_args', nargs='*', help='Optional args (only used by helper scripts)')
args = parser.parse_args()

current_script = Path(argv[0]).name
is_env_var_set = lambda s: len(environ.get(s) or '') > 0
is_html_script = current_script in HTML_SCRIPTS

args.debug = args.deep_debug or args.debug or is_env_var_set('DEBUG')
args.output_emails = args.output_emails or args.all_emails
args.output_other_files = args.output_other_files or args.all_other_files
args.pickled = args.pickled or is_env_var_set('PICKLED') or args.colors_only or len(args.names or []) > 0
args.width = args.width if is_html_script else None
specified_names: list[str | None] = [None if n == 'None' else n for n in (args.names or [])]


# Log level args
if args.deep_debug:
    logger.setLevel(logging.DEBUG)
elif args.debug:
    logger.setLevel(logging.INFO)
elif args.suppress_logs:
    logger.setLevel(logging.FATAL)
elif not env_log_level:
    logger.setLevel(logging.WARNING)

logger.info(f'Log level set to {logger.level}...')
datefinder_logger.setLevel(logger.level)


# Massage args that depend on other args to the appropriate state
if not (args.json_metadata or args.output_texts or args.output_emails or args.output_other_files):
    if is_html_script:
        logger.warning(f"No output section chosen; outputting default of texts, selected emails, and other files...")

    args.output_texts = True
    args.output_emails = True
    args.output_other_files = True

if args.use_epstein_web_links:
    logger.warning(f"Using links to epsteinweb.org links instead of epsteinify.com...")

if args.debug:
    logger.warning(f"Invocation args:\nis_html_script={is_html_script},\nspecified_names={specified_names},\nargs={args}")



import re
from dataclasses import dataclass, field

from rich.highlighter import RegexHighlighter
from rich.text import Text

from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import ARCHIVE_LINK_COLOR
from epstein_files.util.constants import (EMAILER_ID_REGEXES, EPSTEIN_V_ROTHSTEIN_EDWARDS, HEADER_ABBREVIATIONS,
     OSBORNE_LLP, REPLY_REGEX, SENT_FROM_REGEX, VIRGIN_ISLANDS)
from epstein_files.util.doc_cfg import *
from epstein_files.util.data import extract_last_name, listify

CIVIL_ATTORNEY = 'civil attorney'
CRIMINAL_DEFENSE_ATTORNEY = 'criminal defense attorney'
CRIMINAL_DEFENSE_2008 = f"{CRIMINAL_DEFENSE_ATTORNEY} on 2008 case"
EPSTEIN_LAWYER = 'epstein_lawyer'
EPSTEIN_V_ROTHSTEIN_EDWARDS_ATTORNEY = f"{CIVIL_ATTORNEY} working on {EPSTEIN_V_ROTHSTEIN_EDWARDS}"
ESTATE_EXECUTOR = 'estate executor'
EPSTEIN_ESTATE_EXECUTOR = f"Epstein {ESTATE_EXECUTOR}"
REGEX_STYLE_PREFIX = 'regex'
SIMPLE_NAME_REGEX = re.compile(r"^[-\w ]+$", re.IGNORECASE)

CATEGORY_LABEL_MAPPING = {
    ARTICLE: JOURNALIST,
    ARTS: ENTERTAINER,
    BOOK: JOURNALIST,
    LEGAL: EPSTEIN_LAWYER,
    POLITICS: LOBBYIST,
    PROPERTY: BUSINESS,
    REPUTATION: PUBLICIST,
}


@dataclass(kw_only=True)
class HighlightedText:
    """
    Color highlighting for things other than people's names (e.g. phone numbers, email headers).

    Attributes:
        label (str): RegexHighlighter match group name, defaults to 1st 'emailers' key if only 1 emailer provided
        pattern (str): regex pattern identifying strings matching this group
        regex (re.Pattern): matches self.pattern
        style (str): Rich style to apply to text matching this group
        theme_style_name (str): The style name that must be a part of the rich.Console's theme
    """
    label: str = ''
    pattern: str = ''
    style: str
    # Computed fields
    regex: re.Pattern = field(init=False)
    theme_style_name: str = field(init=False)
    _capture_group_label: str = field(init=False)
    _match_group_var: str = field(init=False)

    def __post_init__(self):
        if not self.label:
            raise ValueError(f"No label provided for {repr(self)}")

        self._capture_group_label = self.label.lower().replace(' ', '_').replace('-', '_')
        self._match_group_var = fr"?P<{self._capture_group_label}>"
        self.theme_style_name = f"{REGEX_STYLE_PREFIX}.{self._capture_group_label}"
        self.regex = re.compile(fr"({self._match_group_var}{self.pattern})", re.IGNORECASE | re.MULTILINE)

    def __str__(self) -> str:
        return f"{type(self).__name__}(label='{self.label}')"


@dataclass(kw_only=True)
class HighlightedNames(HighlightedText):
    """
    Encapsulates info about people, places, and other strings we want to highlight with RegexHighlighter.
    Constructor must be called with either an 'emailers' arg or a 'pattern' arg (or both).

    Attributes:
        category (str): optional string to use as an override for self.label in some contexts
        emailers (dict[str, str | None]): optional names to construct regexes for (values are descriptions)
        _pattern (str): complete regex pattern that combines 'pattern' with 'emailers'
    """
    category: str = ''
    emailers: dict[str, str | None] = field(default_factory=dict)
    _pattern: str = field(init=False)

    def __post_init__(self):
        if not (self.emailers or self.pattern):
            raise ValueError(f"Must provide either 'emailers' or 'pattern' arg.")
        elif not self.label:
            if len(self.emailers) == 1:
                self.label = [k for k in self.emailers.keys()][0]
            else:
                raise ValueError(f"No label provided for {repr(self)}")

        super().__post_init__()
        self._pattern = '|'.join([self._emailer_pattern(e) for e in self.emailers] + listify(self.pattern))
        self.regex = re.compile(fr"\b({self._match_group_var}({self._pattern})s?)\b", re.IGNORECASE)

    def get_info(self, name: str) -> str | None:
        """Label and additional info for 'name' if 'name' is in self.emailers."""
        info_pieces = [
            None if len(self.emailers) == 1 else (self.category or self.label.title()),
            self.emailers.get(name),
        ]

        info_pieces = [p for p in info_pieces if p is not None]
        return ', '.join(info_pieces) if info_pieces else None

    def _emailer_pattern(self, name: str) -> str:
        """Pattern matching 'name'. Extends value in EMAILER_ID_REGEXES with last name if it exists."""
        name = remove_question_marks(name)
        last_name = extract_last_name(name)
        first_name = name.removesuffix(f" {last_name}")

        if name in EMAILER_ID_REGEXES:
            pattern = EMAILER_ID_REGEXES[name].pattern

            # Include regex for last name
            # TODO: handle word boundary issue for names that end in symbols
            if SIMPLE_NAME_REGEX.match(last_name) and last_name.lower() not in NAMES_TO_NOT_HIGHLIGHT:
                pattern += fr"|{last_name}"

            return pattern
        elif ' ' not in name:
            return name

        name_patterns = [
            n.replace(' ', r"\s+") for n in [name, first_name, last_name]
            if n.lower() not in NAMES_TO_NOT_HIGHLIGHT
        ]

        return '|'.join(name_patterns)

    def __str__(self) -> str:
        return f"{type(self).__name__}(label='{self.label}')"


HIGHLIGHTED_NAMES = [
    HighlightedNames(
        label='africa',
        style='light_pink4',
        pattern=r'Econet(\s*Wireless)|Ghana(ian)?|Johannesburg|Kenya|Nigerian?|Senegal(ese)?|Serengeti|(South\s*)?African?|(Strive\s*)?Masiyiwa|Tanzania|Ugandan?|Zimbabwe(an)?',
        emailers={
            'Abdoulaye Wade': 'former president of Senegal, father of Karim Wade',
            'Karim Wade': 'son of the president of Senegal, facing arrest for corruption, email handle is "Afri zp"',
            'Miles Alexander': 'Operations Manager Michaelhouse Balgowan KwaZulu-Natal South Africa',  # TODO: what's up with this person?
            'Macky Sall': 'prime minister of Senegal, defeated Abdoulaye Wade',
        },
    ),
    HighlightedNames(
        label='bitcoin',
        style='orange1 bold',
        pattern=r'Balaji|bitcoin|block ?chain(\s*capital)?|Brock(\s*Pierce)?|coins?|cr[iy]?pto(currenc(y|ies))?|e-currency|(Gavin )?Andressen|(Howard\s+)?Lutnic?k|(jeffrey\s+)?wernick|Libra|Madars|(Patrick\s*)?Murck|(Ross\s*)?Ulbricht|Silk\s*Road|SpanCash|Tether|virtual\s*currenc(ies|y)|(zero\s+knowledge\s+|zk)pro(of|tocols?)',
        emailers = {
            JEREMY_RUBIN: 'developer/researcher',
            ANTHONY_SCARAMUCCI: 'Skybridge Capital, FTX investor',
        },
    ),
    HighlightedNames(
        label=BUSINESS,
        style='spring_green4',
        pattern=r'Gruterite|(John\s*)?Kluge|Marc Rich|(Mi(chael|ke)\s*)?Ovitz|(Steve\s+)?Wynn|(Les(lie)?\s+)?Wexner|SALSS|Swedish[-\s]*American\s*Life\s*Science\s*Summit|Valhi|(Yves\s*)?Bouvier',
        emailers = {
            ALIREZA_ITTIHADIEH: 'CEO Freestream Aircraft Limited',
            BARBRO_C_EHNBOM: 'Swedish pharmaceuticals',
            FRED_HADDAD: "co-founder of Heck's in West Virginia",
            GERALD_BARTON: "Maryland property developer, fan of Trump's Irish golf course",
            GORDON_GETTY: 'heir of oil tycoon J. Paul Getty',
            NICHOLAS_RIBIS: 'Hilton CEO, former president of Trump Organization',
            'Philip Kafka': 'president of Prince Concepts (and son of Terry Kafka?)',
            ROBERT_LAWRENCE_KUHN: "investment banker, China expert",
            TERRY_KAFKA: 'CEO of Impact Outdoor (highway billboards)',
            TOM_PRITZKER: 'brother of J.B. Pritzker',
        }
    ),
    HighlightedNames(
        label='cannabis',
        style='chartreuse2',
        pattern=r"CBD|cannabis|marijuana|THC|WEED(guide|maps)?[^s]?",
    ),
    HighlightedNames(
        label='china',
        style='bright_red',
        pattern=r"Ali.?baba|Beijing|CCP|Chin(a|e?se)(?! Daily)|DPRK|Gino\s+Yu|Global Times|Guo|Hong|Huaw[ae]i|Kim\s*Jong\s*Un|Kong|Jack\s+Ma|Kwok|Ministry\sof\sState\sSecurity|Mongolian?|MSS|North\s*Korea|Peking|PRC|SCMP|Tai(pei|wan)|Xi(aomi)?|Jinping",
    ),
    HighlightedNames(
        label='deepak_chopra',
        style='dark_sea_green4',
        emailers = {
            'Carolyn Rangel': 'assistant',
            DEEPAK_CHOPRA: 'woo woo',
        }
    ),
    HighlightedNames(
        label='democrats',
        style='sky_blue1',
        pattern=r'(Al\s*)?Franken|((Bill|Hillart?y)\s*)?Clinton|((Chuck|Charles)\s*)?S(ch|hc)umer|(Diana\s*)?DeGette|DNC|Elena\s*Kagan|(Eliott?\s*)?Spitzer(, Eliot)?|George\s*Mitchell|(George\s*)?Soros|Hill?ary|Dem(ocrat(ic)?)?|(Jo(e|seph)\s*)?Biden|(John\s*)?Kerry|Lisa Monaco|(Matteo\s*)?Salvini|Maxine\s*Waters|(Barac?k )?Obama|(Nancy )?Pelosi|Ron\s*Dellums|Schumer|(Tim\s*)?Geithner|Vernon\s*Jordan',
    ),
    HighlightedNames(
        label='Dubin family',
        style='medium_orchid1',
        pattern=r'((Celina|Eva( Anderss?on)?|Glenn) )?Dubin',
        emailers = {
            GLENN_DUBIN: "Highbridge Capital Management, married to Epstein's ex-gf Eva",
            EVA: "possibly Epstein's ex-girlfriend (?)",
        },
    ),
    HighlightedNames(
        label='employee',
        style='deep_sky_blue4',
        pattern=r'Merwin',
        emailers = {
            'Alfredo Rodriguez': "Epstein's butler, stole the journal",
            ERIC_ROTH: 'jet decorator',
            GWENDOLYN_BECK: 'Epstein fund manager in the 90s',
            JEAN_HUGUEN: 'interior design at Alberto Pinto Cabinet',
            LAWRANCE_VISOSKI: 'pilot',
            LESLEY_GROFF: 'assistant',
            'Linda Pinto': 'interior design at Alberto Pinto Cabinet',
            MERWIN_DELA_CRUZ: None,  # HOUSE_OVERSIGHT_032652 Groff says "Jojo and Merwin both requested off Nov. 25 and 26"
            NADIA_MARCINKO: 'pilot',
        }
    ),
    HighlightedNames(
        label=ENTERTAINER,
        style='light_steel_blue3',
        pattern=r'(Art )?Spiegelman|Bobby slayton|bono\s*mick|Errol(\s*Morris)?|Etienne Binant|(Frank\s)?Gehry|Jagger|(Jeffrey\s*)?Katzenberg|(Johnny\s*)?Depp|Kid Rock|Lena\s*Dunham|Madonna|Mark\s*Burnett|Ramsey Elkholy|shirley maclaine|Steven Gaydos?|Woody( Allen)?|Zach Braff',
        emailers={
            ANDRES_SERRANO: "'Piss Christ' artist",
            'Barry Josephson': 'American film producer and former music manager, editor FamilySecurityMatters.org',
            BILL_SIEGEL: 'documentary film producer and director',
            DAVID_BLAINE: 'famous magician',
            HENRY_HOLT: f"{MICHAEL_WOLFF}'s book publisher",
            'Richard Merkin': 'painter, illustrator and arts educator',
            STEVEN_PFEIFFER: 'Associate Director at Independent Filmmaker Project (IFP)',
        },
    ),
    HighlightedNames(
        label=EPSTEIN_LAWYER,
        style='purple',
        pattern=r'(Barry (E. )?)?Krischer|Kate Kelly|Kirkland\s*&\s*Ellis|(Leon\s*)?Jaworski|Michael J. Pike|Paul,?\s*Weiss|Steptoe|Wein(berg|garten)',
        emailers = {
            'Alan S Halperin': 'parnter at Paul, Weiss',
            ARDA_BESKARDES: 'NYC immigration attorney allegedly involved in sex-trafficking operations',
            BENNET_MOSKOWITZ: f'represented the {EPSTEIN_ESTATE_EXECUTOR}s',
            BRAD_KARP: 'head of the law firm Paul Weiss',
            DAVID_SCHOEN: f"{CRIMINAL_DEFENSE_ATTORNEY} after 2019 arrest",
            DEBBIE_FEIN: EPSTEIN_V_ROTHSTEIN_EDWARDS_ATTORNEY,
            'Erika Kellerhals': 'attorney in St. Thomas',
            GERALD_LEFCOURT: f'friend of {ALAN_DERSHOWITZ}',
            JACK_GOLDBERGER: CRIMINAL_DEFENSE_2008,
            JACKIE_PERCZEK: CRIMINAL_DEFENSE_2008,
            JAY_LEFKOWITZ: f"Kirkland & Ellis partner, {CRIMINAL_DEFENSE_2008}",
            JESSICA_CADWELL: 'paralegal',  # paralegal, see https://x.com/ImDrinknWyn/status/1993765348898927022
            LILLY_SANCHEZ: CRIMINAL_DEFENSE_ATTORNEY,
            MARTIN_WEINBERG: CRIMINAL_DEFENSE_ATTORNEY,
            MICHAEL_MILLER: 'Steptoe LLP partner',
            REID_WEINGARTEN: 'Steptoe LLP partner',
            'Roy Black': CRIMINAL_DEFENSE_2008,
            SCOTT_J_LINK: None,
            TONJA_HADDAD_COLEMAN: f'{EPSTEIN_V_ROTHSTEIN_EDWARDS_ATTORNEY}, maybe daughter of Fred Haddad?',
        }
    ),
    HighlightedNames(
        label=ESTATE_EXECUTOR,
        style='purple3 bold',
        category='epstein lawyer',
        emailers = {
            DARREN_INDYKE: EPSTEIN_ESTATE_EXECUTOR,
            RICHARD_KAHN: EPSTEIN_ESTATE_EXECUTOR,
        }
    ),
    HighlightedNames(
        label='europe',
        style='light_sky_blue3',
        pattern=r'(Angela )?Merk(el|le)|Austria|(Benjamin\s*)?Harnwell|Berlin|Brexit(eers?)?|Brit(ain|ish)|Brussels|Cannes|(Caroline|Jack)?\s*Lang(, Caroline)?|Cypr(iot|us)|Davos|ECB|EU|Europe(an)?(\s*Union)?|France|Geneva|Germany?|Gillard|Gree(ce|k)|Ital(ian|y)|Jacques|(Kevin\s*)?Rudd|Le\s*Pen|London|Macron|Melusine|Munich|(Natalia\s*)?Veselnitskaya|(Nicholas\s*)?Sarkozy|Nigel(\s*Farage)?|Oslo|Paris|Polish|(Sebastian )?Kurz|(Vi(c|k)tor\s+)?Orbah?n|Edward Rod Larsen|Strasbourg|Strauss[- ]?Kahn|Swed(en|ish)(?![-\s]+America)|Switzerland|(Tony\s)?Blair|Ukrain(e|ian)|Vienna|(Vitaly\s*)?Churkin|Zug',
        emailers = {
            ANDRZEJ_DUDA: 'former president of Poland',
            MIROSLAV_LAJCAK: 'Russia-friendly Slovakian politician, friend of Steve Bannon',
            PETER_MANDELSON: 'UK politics',
            TERJE_ROD_LARSEN: 'Norwegian diplomat',
            THORBJORN_JAGLAND: 'former prime minister of Norway and head of the Nobel Peace Prize Committee',
        }
    ),
    HighlightedNames(
        label='famous_lawyer',
        style='medium_purple3',
        category='famous_lawyer',
        pattern=r'(David\s*)?Bo[il]es|dersh|(Gloria\s*)?Allred|(Mi(chael|ke)\s*)?Avenatti',
        emailers = {
            ALAN_DERSHOWITZ: 'Harvard Law School professor and all around (in)famous American lawyer',
            KEN_STARR: 'head of the Monica Lewinsky investigation against Bill Clinton',
        }
    ),
    HighlightedNames(
        label='friend',
        style='tan',
        pattern=r"Andrew Farkas|Thomas\s*(J\.?\s*)?Barrack(\s*Jr)?",
        emailers = {
            DAVID_STERN: f'emailed Epstein from Moscow, appears to know chairman of {DEUTSCHE_BANK}',
            JONATHAN_FARKAS: "heir to the Alexander's department store fortune",
            'Peter Thomas Roth': 'student of Epstein at Dalton, skincare company founder',
            STEPHEN_HANSON: None,
            TOM_BARRACK: 'long time friend of Trump',
        }
    ),
    HighlightedNames(
        label='finance',
        style='green',
        pattern=r'Apollo|Ari\s*Glass|(Bernie\s*)?Madoff|Black(rock|stone)|BofA|Boothbay(\sFund\sManagement)?|Chase\s*Bank|Credit\s*Suisse|DB|Deutsche\s*(Asset|Bank)|Electron\s*Capital\s*(Partners)?|Fenner|FRBNY|Goldman(\s*Sachs)|HSBC|Invesco|(Janet\s*)?Yellen|(Jerome\s*)?Powell(?!M\. Cabot)|(Jimmy\s*)?Cayne|JPMC?|j\.?p\.?\s*morgan(\.?com|\s*Chase)?|Madoff|Merrill(\s*Lynch)?|(Michael\s*)?(Cembalest|Milken)|MLPF&S|(money\s+)?launder(s?|ers?|ing)?(\s+money)?|Morgan Stanley|(Peter L. )?Scher|(Ray\s*)?Dalio|Schwartz?man|Serageldin|UBS|us.gio@jpmorgan.com',
        emailers={
            AMANDA_ENS: 'Citigroup',
            DANIEL_SABBA: 'UBS Investment Bank',
            DAVID_FISZEL: 'CIO Honeycomb Asset Management',
            JES_STALEY: 'former CEO of Barclays',
            JIDE_ZEITLIN: 'former partner at Goldman Sachs, allegations of sexual misconduct',
            LEON_BLACK: 'Apollo CEO',
            MARC_LEON: 'Luxury Properties Sari Morrocco',
            MELANIE_SPINELLA: f'representative of {LEON_BLACK}',
            MORTIMER_ZUCKERMAN: 'business partner of Epstein',
            PAUL_BARRETT: None,
            PAUL_MORRIS: 'Deutsche Bank',
        }
    ),
    HighlightedNames(
        label=HARVARD.lower(),
        style='deep_pink2',
        pattern=r'Cambridge|(Derek\s*)?Bok|Elisa(\s*New)?|Harvard(\s*(Business|Law|University)(\s*School)?)?|(Jonathan\s*)?Zittrain|(Stephen\s*)?Kosslyn',
        emailers = {
            "Kelly Friendly": f"longtime aide and spokesperson of {LARRY_SUMMERS}",
            LARRY_SUMMERS: 'board of Digital Currency Group (DCG), Harvard president, Obama economic advisor',
            'Leah Reis-Dennis': 'producer for Lisa New\'s Poetry in America',
            LISA_NEW: f'professor of poetry, wife of {LARRY_SUMMERS}, AKA "Elisa New"',
            'Lisa Randall': 'theoretical physicist',
            MARTIN_NOWAK: 'professor of mathematics and biology',
            MOSHE_HOFFMAN: 'lecturer and research scholar in behavioral and evolutionary economics',
        }
    ),
    HighlightedNames(
        label='india',
        style='bright_green',
        pattern=r'Abraaj|Anna\s*Hazare|(Arif\s*)?Naqvi|(Arvind\s*)?Kejriwal|Hardeep( Pur[ei]e)?|Indian?|InsightsPod|Modi|Mumbai|Tranchulas',
        emailers = {
            ANIL_AMBANI: 'chairman of Reliance Group',
            VINIT_SAHNI: None,
            ZUBAIR_KHAN: 'Tranchulas CEO, InsightsPod founder',
        }
    ),
    HighlightedNames(
        label='israel',
        style='dodger_blue2',
        pattern=r"AIPAC|Bibi|(eh|(Ehud|Nili Priell) )?barak|Ehud\s*Barack|Israeli?|Jerusalem|J\s*Street|Mossad|Netanyahu|(Sheldon\s*)?Adelson|Tel\s*Aviv|(The\s*)?Shimon\s*Post|Yitzhak|Rabin|YIVO|zionist",
        emailers={
            EHUD_BARAK: 'former primer minister',
            'Mitchell Bard': 'director of the American-Israeli Cooperative Enterprise (AICE)',
            'Nili Priell Barak': f'wife of {EHUD_BARAK}',
        }
    ),
    HighlightedNames(
        label='japan',
        style='color(168)',
        pattern=r'BOJ|(Bank\s+of\s+)?Japan(ese)?|jpy?(?! Morgan)|SG|Singapore|Toky[op]',
    ),
    HighlightedNames(
        label='javanka',
        style='medium_violet_red',
        emailers = {
            IVANKA: None,
            JARED_KUSHNER: None,
        }
    ),
    HighlightedNames(
        label=JOURNALIST,
        style='bright_yellow',
        pattern=r'Palm\s*Beach\s*(Daily\s*News|Post)|ABC(\s*News)?|Alex\s*Yablon|(Andrew\s*)?Marra|Arianna(\s*Huffington)?|(Arthur\s*)?Kretchmer|BBC|Bloomberg|Breitbart|Charlie\s*Rose|China\s*Daily|CNBC|CNN(politics?)?|Con[cs]hita|Sarnoff|(?<!Virgin[-\s]Islands[-\s])Daily\s*(Beast|Mail|News|Telegraph)|(David\s*)?Pecker|David\s*Brooks|Ed\s*Krassenstein|(Emily\s*)?Michot|Ezra\s*Klein|(George\s*)?Stephanopoulus|Globe\s*and\s*Mail|Good\s*Morning\s*America|Graydon(\s*Carter)?|Huffington(\s*Post)?|Ingram, David|(James\s*)?Patterson|Jonathan\s*Karl|Julie\s*(K.?\s*)?Brown|(Katie\s*)?Couric|Keith\s*Larsen|L\.?A\.?\s*Times|Miami\s*Herald|(Michele\s*)?Dargan|(National\s*)?Enquirer|(The\s*)?N(ew\s*)?Y(ork\s*)?(P(ost)?|T(imes)?)|(The\s*)?New\s*Yorker|NYer|PERVERSION\s*OF\s*JUSTICE|Politico|Pro\s*Publica|(Sean\s*)?Hannity|Sulzberger|SunSentinel|Susan Edelman|(Uma\s*)?Sanghvi|(The\s*)?Wa(shington\s*)?Po(st)?|Viceland|Vick[iy]\s*Ward|Vox|WGBH|(The\s*)?Wall\s*Street\s*Journal|WSJ|[-\w.]+@(bbc|independent|mailonline|mirror|thetimes)\.co\.uk',
        emailers = {
            EDWARD_JAY_EPSTEIN: 'reporter who wrote about the kinds of crimes Epstein was involved in, no relation to Jeffrey',
            'James Hill': 'ABC News',
            JENNIFER_JACQUET: 'Future Science',
            JOHN_BROCKMAN: 'literary agent and author specializing in scientific literature',
            LANDON_THOMAS: 'New York Times',
            MICHAEL_WOLFF: 'Author of "Fire and Fury: Inside the Trump White House"',
            PAUL_KRASSNER: '60s counterculture guy',
            'Tim Zagat': 'Zagat restaurant guide CEO',
        }
    ),
    HighlightedNames(
        label='latin america',
        style='yellow',
        pattern=r'Argentin(a|ian)|Bolsonar[aio]|Bra[sz]il(ian)?|Bukele|Caracas|Castro|Colombian?|Cuban?|El\s*Salvador|((Enrique )?Pena )?Nieto|LatAm|Lula|Mexic(an|o)|(Nicolas\s+)?Maduro|Panama( Papers)?|Peru|Venezuelan?|Zambrano',
    ),
    HighlightedNames(
        label='law enforcement',
        style='color(24) bold',
        pattern=r'ag|(Alicia\s*)?Valle|((Bob|Robert)\s*)?Mueller|(Byung\s)?Pak|CFTC|CIA|CIS|CVRA|Dep(artmen)?t\.?\s*of\s*(the\s*)?(Justice|Treasury)|DHS|DOJ|FBI|FCPA|FDIC|Federal\s*Bureau\s*of\s*Investigation|FinCEN|FINRA|FOIA|FTC|IRS|(James\s*)?Comey|(Jennifer\s*Shasky\s*)?Calvery|((Judge|Mark)\s*)?(Carney|Filip)|(Kirk )?Blouin|KYC|NIH|NS(A|C)|OCC|OFAC|(Lann?a\s*)?Belohlavek|(Michael\s*)?Reiter|OGE|Office\s*of\s*Government\s*Ethics|Police Code Enforcement|(Preet\s*)?Bharara|SCOTUS|SD(FL|NY)|Southern\s*District\s*of\s*(Florida|New\s*York)|SEC|Securities\s*and\s*Exchange\s*Commission|State\s*Dep(artmen)?t|Strzok|Supreme\s*Court|Treasury\s*(Dep(artmen)?t|Secretary)|TSA|USAID|(William\s*J\.?\s*)?Zloch',
        emailers = {
            ANN_MARIE_VILLAFANA: 'southern district of Florida U.S. Attorney',
            DANNY_FROST: 'Director of Communications at Manhattan DA',
        }
    ),
    HighlightedNames(
        label=LOBBYIST,
        style='light_coral',
        pattern=r'[BR]ob Crowe|Stanley Rosenberg',
        emailers = {
            'Joshua Cooper Ramo': 'co-CEO of Henry Kissinger Associates',
            KATHERINE_KEATING: 'Daughter of former Australian PM',
            MOHAMED_WAHEED_HASSAN: 'former president of the Maldives',
            OLIVIER_COLOM: 'France',
            'Paul Keating': 'former PM of Australia',
            PUREVSUREN_LUNDEG: 'Mongolian ambassador to the UN',
            'Stanley Rosenberg': 'former President of the Massachusetts Senate',
        }
    ),
    HighlightedNames(
        label='mideast',
        style='dark_sea_green4',
        # this won't match ever because of word boundary: [-\s]9/11[\s.]
        pattern=r"Abdulmalik Al-Makhlafi|Abdullah|Abu\s+Dhabi|Afghanistan|Al[-\s]?Qa[ei]da|Ahmadinejad|Arab|Aramco|Assad|Bahrain|Basiji?|Benghazi|Cairo|Chagoury|Dj[iu]bo?uti|Doha|Dubai|Egypt(ian)?|Emir(at(es?|i))?|Erdogan|Fashi|Gaddafi|(Hamid\s*)?Karzai|Hamad\s*bin\s*Jassim|HBJ|Houthi|Imran\s+Khan|Iran(ian)?|Isi[ls]|Islam(abad|ic|ist)?|Istanbul|Kh?ashoggi|(Kairat\s*)?Kelimbetov|kasshohgi|Kaz(akh|ich)stan|Kazakh?|Kh[ao]menei|Khalid\s*Sheikh\s*Mohammed|KSA|Leban(ese|on)|Libyan?|Mahmoud|Marra[hk]e[cs]h|MB(N|S|Z)|Mohammed\s+bin\s+Salman|Morocco|Mubarak|Muslim|Nayaf|Pakistani?|Omar|(Osama\s*)?Bin\s*Laden|Osama(?! al)|Palestin(e|ian)|Persian?|Riya(dh|nd)|Saddam|Salman|Saudi(\s+Arabian?)?|Shariah?|SHC|sheikh|shia|(Sultan\s*)?Yacoub|Syrian?|(Tarek\s*)?El\s*Sayed|Tehran|Tunisian?|Turk(ey|ish)|UAE|((Iraq|Iran|Kuwait|Qatar|Yemen)i?)",
        emailers = {
            ANAS_ALRASHEED: f'former information minister of Kuwait {QUESTION_MARKS}',
            AZIZA_ALAHMADI: 'Abu Dhabi Department of Culture & Tourism',
            RAAFAT_ALSABBAGH: 'Saudi royal advisor',
            SHAHER_ABDULHAK_BESHER: 'Yemeni billionaire',
        }
    ),
    HighlightedNames(
        label='modeling',
        style='pale_violet_red1',
        pattern=r'\w+@mc2mm.com|(Nicole\s*)?Junkerman',
        emailers = {
            'Abi Schwinck': 'MC2 Model Management (?)',
            DANIEL_SIAD: None,
            FAITH_KATES: 'Next Models co-founder',
            'Gianni Serazzi': 'fashion consultant?',
            HEATHER_MANN: 'South African former model, ex-girlfriend of Prince Andrew (?)',
            JEAN_LUC_BRUNEL: 'MC2 Model Management founder, died by suicide in French jail',
            JEFF_FULLER: 'president of MC2 Model Management USA',
            MANUELA_MARTINEZ: 'Mega Partners (Brazilian agency)',
            MARIANA_IDZKOWSKA: None,
            'Michael Sanka': 'MC2 Model Management (?)',
        }
    ),
    HighlightedNames(
        label=PUBLICIST,
        style='orange_red1',
        pattern=fr"(Matt(hew)? )?Hiltzi[gk]|{REPUTATION_MGMT}",
        emailers = {
            AL_SECKEL: 'husband of Isabel Maxwell, Mindshift conference organizer who fell off a cliff',
            'Barnaby Marsh': 'co-founder of Saint Partners, a philanthropy services company',
            CHRISTINA_GALBRAITH: f"{REPUTATION_MGMT}, worked on Epstein's Google search results with {TYLER_SHEARS}",
            IAN_OSBORNE: f"{OSBORNE_LLP} reputation repairer possibly hired by Epstein ca. 2011-06",
            MICHAEL_SITRICK: 'crisis PR',
            PEGGY_SIEGAL: 'socialite',
            ROSS_GOW: 'Acuity Reputation Management',
            TYLER_SHEARS: f"{REPUTATION_MGMT}, worked on Epstein's Google search results with {CHRISTINA_GALBRAITH}",
        }
    ),
    HighlightedNames(
        label='republicans',
        style='bold dark_red',
        pattern=r'Alberto\sGonzale[sz]|(Alex\s*)?Acosta|(Bill\s*)?Barr|Bill\s*Shine|(Bob\s*)?Corker|(John\s*(R.?\s*)?)Bolton|Broidy|(Chris\s)?Christie|Devin\s*Nunes|(Don\s*)?McGa[hn]n|McMaster|(George\s*)?Nader|GOP|(Brett\s*)?Kavanaugh|Kissinger|Kobach|Koch\s*Brothers|Kolfage|Kudlow|Lewandowski|(Marco\s)?Rubio|(Mark\s*)Meadows|Mattis|(?<!Merwin Dela )Cruz|(Michael\s)?Hayden|((General|Mike)\s*)?(Flynn|Pence)|(Mitt\s*)?Romney|Mnuchin|Nikki|Haley|(Paul\s+)?Manafort|(Peter\s)?Navarro|Pompeo|Reagan|Republican|(?<!Cynthia )(Richard\s*)?Nixon|Sasse|(Rex\s*)?Tillerson',
        emailers = {
            RUDY_GIULIANI: 'disbarred formed mayor of New York City',
            TULSI_GABBARD: None,
        },
    ),
    HighlightedNames(
        label='Rothschild family',
        style='indian_red',
        emailers={
            ARIANE_DE_ROTHSCHILD: None,
            JOHNNY_EL_HACHEM: f'Works with {ARIANE_DE_ROTHSCHILD}',
        },
    ),
    HighlightedNames(
        label='russia',
        style='red bold',
        pattern=r'Alfa\s*Bank|Anya\s*Rasulova|Chernobyl|Day\s+One\s+Ventures|(Dmitry\s)?(Kiselyov|(Lana\s*)?Pozhidaeva|Medvedev|Rybolo(o?l?ev|vlev))|Dmitry|FSB|GRU|KGB|Kislyak|Kremlin|Kuznetsova|Lavrov|Lukoil|Moscow|(Oleg\s*)?Deripaska|Oleksandr Vilkul|Rosneft|RT|St.?\s*?Petersburg|Russian?|Sberbank|Soviet(\s*Union)?|USSR|(Vladimir\s*)?(Putin|Yudashkin)|Women\s*Empowerment|Xitrans',
        emailers = {
            MASHA_DROKOVA: 'silicon valley VC, former Putin Youth',
            RENATA_BOLOTOVA: 'former aspiring model, now fund manager at New York State Insurance Fund',
            SVETLANA_POZHIDAEVA: f'Epstein\'s Russian assistant who was recommended for a visa by Sergei Belyakov (FSB) and {DAVID_BLAINE}',
        }
    ),
    HighlightedNames(
        label=ACADEMIA,
        style='light_goldenrod2',
        pattern=r'Alain Forget|Brotherton|Carl\s*Sagan|Columbia|David Grosof|J(ames|im)\s*Watson|(Lord\s*)?Martin\s*Rees|Massachusetts\s*Institute\s*of\s*Technology|MIT(\s*Media\s*Lab)?|Media\s*Lab|Minsky|((Noam|Valeria)\s*)?Chomsky|Praluent|Regeneron|(Richard\s*)?Dawkins|Sanofi|Stanford|(Stephen\s*)?Hawking|(Steven?\s*)?Pinker|UCLA',
        emailers = {
            DAVID_HAIG: None,
            JOSCHA_BACH: 'cognitive science / AI research',
            'Daniel Kahneman': 'Nobel economic sciences laureate and cognitivie psychologist (?)',
            LAWRENCE_KRAUSS: 'theoretical physicist',
            LINDA_STONE: 'ex-Microsoft, MIT Media Lab',
            MARK_TRAMO: 'professor of neurology at UCLA',
            NEAL_KASSELL: 'professor of neurosurgery at University of Virginia',
            PETER_ATTIA: 'longevity medicine',
            ROBERT_TRIVERS: 'evolutionary biology',
            ROGER_SCHANK: 'Teachers College, Columbia University',
        },
    ),
    HighlightedNames(
        label='southeast_asia',
        style='light_salmon3 bold',
        pattern=r'Bangkok|Burm(a|ese)|Cambodian?|Laos|Malaysian?|Myan?mar|Thai(land)?|Vietnam(ese)?',
    ),
    HighlightedNames(
        label='tech bro',
        style='bright_cyan',
        pattern=r"AG?I|Chamath|Palihapitiya|Danny\s*Hillis|Drew\s*Houston|Eric\s*Schmidt|Greylock(\s*Partners)?|(?<!(ustin|Moshe)\s)Hoffmand?|LinkedIn|(Mark\s*)?Zuckerberg|Masa(yoshi)?(\sSon)?|Najeev|Nathan\s*Myhrvold|Palantir|(Peter\s)?Th(ie|ei)l|Pierre\s*Omidyar|Sergey\s*Brin|Silicon\s*Valley|Softbank|SpaceX|Tim\s*Ferriss?|WikiLeak(ed|s)",
        emailers = {
            'Auren Hoffman': 'CEO of SafeGraph (firm that gathers location data from mobile devices) and LiveRamp',
            ELON_MUSK: 'father of Mecha-Hitler',
            PETER_THIEL: 'Paypal mafia member, founder of Palantir, early Facebook investor, reactionary',
            REID_HOFFMAN: 'PayPal mafia member, founder of LinkedIn',
            STEVEN_SINOFSKY: 'ex-Microsoft, loves bitcoin',
        },
    ),
    HighlightedNames(
        label='trump',
        style='red3 bold',
        pattern=r"@?realDonaldTrump|(Alan\s*)?Weiss?elberg|\bDJ?T\b|Donald J. Tramp|(Donald\s+(J\.\s+)?)?Trump(ism|\s*Properties)?|Don(ald| *Jr)(?! Rubin)|Ivana|(Madeleine\s*)?Westerhout|Mar[-\s]*a[-\s]*Lago|(Marla\s*)?Maples|(Matt(hew)? )?Calamari|\bMatt C\b|Melania|(Michael (J.? )?)?Boccio|Roger\s+Stone|rona|(The\s*)?Art\s*of\s*the\s*Deal",
        emailers = {
            'Bruce Moskowitz': "'Trump's health guy' according to Epstein",
        },
    ),
    HighlightedNames(
        label='victim',
        style='orchid1',
        pattern=r'BVI|(Jane|Tiffany)\s*Doe|Katie\s*Johnson|(Virginia\s+((L\.?|Roberts)\s+)?)?Giuffre|Virginia\s+Roberts',
    ),
    HighlightedNames(
        label='victim_lawyer',
        style='dark_magenta bold',
        pattern=r'(Alan(\s*P.)?|MINTZ)\s*FRAADE|Paul\s*(G.\s*)?Cassell|Rothstein\s*Rosenfeldt\s*Adler|(Scott\s*)?Rothstein|(J\.?\s*)?(Stan(ley)?\s*)?Pottinger',
        emailers = {
            BRAD_EDWARDS: 'Rothstein Rosenfeldt Adler (Rothstein was a crook & partner of Roger Stone)',
            JACK_SCAROLA: 'Searcy Denney Scarola Barnhart & Shipley',
        }
    ),
    HighlightedNames(
        label=VIRGIN_ISLANDS,
        style='sea_green1',
        pattern=r'Bahamas|Caribb?ean|Dominican\s*Republic|(Great|Little)\s*St.?\s*James|Haiti(an)?|(John\s*)deJongh(\s*Jr\.?)|(Kenneth E\. )?Mapp|Palm\s*Beach(?!\s*Post)|PBI|S(ain)?t.?\s*Thomas|USVI|VI|(The\s*)?Virgin\s*Islands(\s*Daily\s*News)?',  # TODO: VI Daily News should be yellow but it's hard bc Daily News xists
        emailers = {
            CECILE_DE_JONGH: f'First lady 2007-2015',
            STACEY_PLASKETT: 'non-voting member of Congress',
            KENNETH_E_MAPP: 'Governor',
        },
    ),

    # Individuals
    HighlightedNames(
        label=BILL_GATES,
        style='turquoise4',
        pattern=r'BG|b?g?C3|(Bill\s*((and|or)\s*Melinda\s*)?)?Gates|Melinda(\s*Gates)?|Microsoft|MSFT',
        emailers = {
            BORIS_NIKOLIC: f'biotech VC partner of {BILL_GATES}, {EPSTEIN_ESTATE_EXECUTOR}',
        },
    ),
    HighlightedNames(
        label=STEVE_BANNON,
        style='color(58)',
        pattern=r'((Steve|Sean)\s*)?Bannon?',
    ),
    HighlightedNames(
        emailers={STEVEN_HOFFENBERG: HEADER_ABBREVIATIONS['Hoffenberg']},
        pattern=r'(steven?\s*)?hoffenberg?w?',
        style='gold3'
    ),
    HighlightedNames(emailers={GHISLAINE_MAXWELL: None}, pattern='gmax(1@ellmax.com)?|TerraMar', style='deep_pink3'),
    HighlightedNames(emailers={JABOR_Y: HEADER_ABBREVIATIONS['Jabor']}, style='spring_green1'),
    HighlightedNames(emailers={JEFFREY_EPSTEIN: None}, pattern='JEGE|LSJ|Mark (L. )?Epstein', style='blue1'),
    HighlightedNames(emailers={JOI_ITO: 'former head of MIT Media Lab'}, style='gold1'),
    HighlightedNames(emailers={KATHRYN_RUEMMLER: 'former Obama legal counsel'}, style='magenta2'),
    HighlightedNames(emailers={MELANIE_WALKER: 'doctor'}, style='pale_violet_red1'),
    HighlightedNames(emailers={PAULA: "Epstein's ex-girlfriend who is now in the opera"}, label='paula_heil_fisher', style='pink1'),
    HighlightedNames(emailers={PRINCE_ANDREW: 'British royal family'}, style='dodger_blue1'),
    HighlightedNames(emailers={SOON_YI_PREVIN: "wife of Woody Allen"}, style='hot_pink'),
    HighlightedNames(emailers={SULTAN_BIN_SULAYEM: 'CEO of DP World, chairman of ports in Dubai'}, style='green1'),
    HighlightedText(label='unknown', style='cyan', pattern=r'\(unknown\)'),  # HighlightedText bc of word boundary issue
]

# Highlight regexes for things other than names, only used by RegexHighlighter pattern matching
HIGHLIGHTED_TEXTS = [
    HighlightedText(
        label='header_field',
        style='plum4',
        pattern=r'^(Date|From|Sent|To|C[cC]|Importance|Subject|Bee|B[cC]{2}|Attachments):',
    ),
    HighlightedText(
        label='http_links',
        style=f'{ARCHIVE_LINK_COLOR} underline',
        pattern=r"https?:[^\s]+",
    ),
    HighlightedText(
        label='phone_number',
        style='bright_green',
        pattern=r"\+?(1?\(?\d{3}\)?[- ]\d{3}[- ]\d{4}|\d{2}[- ]\(?0?\)?\d{2}[- ]\d{4}[- ]\d{4})|[\d+]{10,12}",
    ),
    HighlightedText(
        label='quoted_reply_line',
        style='dim',
        pattern=REPLY_REGEX.pattern,
    ),
    HighlightedText(
        label='redacted',
        style='grey58',
        pattern=fr"{REDACTED}|Privileged - Redacted",
    ),
    HighlightedText(
        label='sent_from',
        style='gray42 italic',
        pattern=SENT_FROM_REGEX.pattern,
    ),
    HighlightedText(
        label='snipped_signature',
        style='gray19',
        pattern=r'<\.\.\.(snipped|trimmed).*\.\.\.>',
    ),
    HighlightedText(
        label='timestamp_2',
        style=TIMESTAMP_STYLE,
        pattern=r"\d{1,4}[-/]\d{1,2}[-/]\d{2,4} \d{1,2}:\d{2}:\d{2}( [AP]M)?",
    ),
]

ALL_HIGHLIGHTS = HIGHLIGHTED_NAMES + HIGHLIGHTED_TEXTS


class EpsteinHighlighter(RegexHighlighter):
    """rich.highlighter that finds and colors interesting keywords based on the above config."""
    base_style = f"{REGEX_STYLE_PREFIX}."
    highlights = [highlight_group.regex for highlight_group in ALL_HIGHLIGHTS]


def get_info_for_name(name: str) -> str | None:
    highlight_group = _get_highlight_group_for_name(name)

    if highlight_group and isinstance(highlight_group, HighlightedNames):
        return highlight_group.get_info(name)


def get_style_for_category(category: str) -> str | None:
    if category in [CONFERENCE, SPEECH]:
        return f"{get_style_for_category(ACADEMIA)} dim"
    elif category == JSON:
        return 'dark_red'
    elif category == JUNK:
        return 'grey19'
    elif category == 'letter':
        return 'medium_orchid1'
    elif category == SOCIAL:
        return f"{get_style_for_category(PUBLICIST)} dim"

    category = CATEGORY_LABEL_MAPPING.get(category, category)

    for highlight_group in HIGHLIGHTED_NAMES:
        if highlight_group.label == category:
            return highlight_group.style


def get_style_for_name(name: str | None, default_style: str = DEFAULT, allow_bold: bool = True) -> str:
    highlight_group = _get_highlight_group_for_name(name or UNKNOWN)
    style = highlight_group.style if highlight_group else default_style
    return style if allow_bold else style.replace('bold', '').strip()


def _get_highlight_group_for_name(name: str) -> HighlightedNames | None:
    for highlight_group in HIGHLIGHTED_NAMES:
        if highlight_group.regex.search(name):
            return highlight_group



# Rich reference: https://rich.readthedocs.io/en/latest/reference.html
import json
from os import devnull
from pathlib import Path
from typing import Literal

from rich.align import Align
from rich.console import Console, RenderableType
from rich.markup import escape
from rich.panel import Panel
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.theme import Theme

from epstein_files.util.constant.html import CONSOLE_HTML_FORMAT, HTML_TERMINAL_THEME, PAGE_TITLE
from epstein_files.util.constant.names import UNKNOWN
from epstein_files.util.constant.strings import DEFAULT, EMAIL, NA, OTHER_SITE_LINK_STYLE, QUESTION_MARKS, SiteType
from epstein_files.util.constant.urls import *
from epstein_files.util.constants import FALLBACK_TIMESTAMP, HEADER_ABBREVIATIONS
from epstein_files.util.data import json_safe
from epstein_files.util.env import args
from epstein_files.util.file_helper import file_size_str
from epstein_files.util.highlighted_group import ALL_HIGHLIGHTS, HIGHLIGHTED_NAMES, EpsteinHighlighter
from epstein_files.util.logging import logger

TITLE_WIDTH = 50
NUM_COLOR_KEY_COLS = 4
NA_TXT = Text(NA, style='dim')
QUESTION_MARK_TXT = Text(QUESTION_MARKS, style='dim')
GREY_NUMBERS = [58, 39, 39, 35, 30, 27, 23, 23, 19, 19, 15, 15, 15]

DEFAULT_NAME_STYLE = 'gray46'
KEY_STYLE='honeydew2 bold'
SECTION_HEADER_STYLE = 'bold white on blue3'
SOCIAL_MEDIA_LINK_STYLE = 'cyan3 bold'
SUBSTACK_POST_LINK_STYLE = 'bright_cyan'
SYMBOL_STYLE = 'grey70'
TITLE_STYLE = 'black on bright_white bold'

HIGHLIGHTED_GROUP_COLOR_KEYS = [
    Text(highlight_group.label.replace('_', ' '), style=highlight_group.style)
    for highlight_group in sorted(HIGHLIGHTED_NAMES, key=lambda hg: hg.label)
]

THEME_STYLES = {
    DEFAULT: 'wheat4',
    TEXT_LINK: 'deep_sky_blue4 underline',
    **{hg.theme_style_name: hg.style for hg in ALL_HIGHLIGHTS},
}

# Instantiate console object
CONSOLE_ARGS = {
    'color_system': '256',
    'highlighter': EpsteinHighlighter(),
    'record': args.build,
    'safe_box': False,
    'theme': Theme(THEME_STYLES),
    'width': args.width,
}

if args.suppress_output:
    logger.warning(f"Suppressing terminal output because args.suppress_output={args.suppress_output}...")
    CONSOLE_ARGS.update({'file': open(devnull, "wt")})

console = Console(**CONSOLE_ARGS)
highlighter = CONSOLE_ARGS['highlighter']


def add_cols_to_table(table: Table, col_names: list[str]) -> None:
    """Left most col will be left justified, rest are center justified."""
    for i, col in enumerate(col_names):
        table.add_column(col, justify='left' if i == 0 else 'center')


def join_texts(txts: list[Text], join: str = ' ', encloser: str = '') -> Text:
    """Join rich.Text objs into one."""
    if encloser:
        if len(encloser) != 2:
            raise ValueError(f"'encloser' arg is '{encloser}' which is not 2 characters long")

        enclose_start, enclose_end = (encloser[0], encloser[1])
    else:
        enclose_start = enclose_end = ''

    txt = Text('')

    for i, link in enumerate(txts):
        txt.append(join if i >= 1 else '').append(enclose_start).append(link).append(enclose_end)

    return txt


def key_value_txt(key: str, value: Text | str) -> Text:
    """Generate a Text obj for 'key=value'."""
    return Text('').append(key, style=KEY_STYLE).append('=', style=SYMBOL_STYLE).append(value)


def parenthesize(msg: str | Text, style: str = '') -> Text:
    txt = Text(msg) if isinstance(msg, str) else msg
    return Text('(', style=style).append(txt).append(')')


def print_author_header(msg: str, color: str | None, footer: str | None = None) -> None:
    txt = Text(msg, justify='center')
    color = color or 'white'
    color = 'white' if color == DEFAULT else color
    panel = Panel(txt, width=80, style=f"black on {color} bold")
    console.print('\n', Align.center(panel))

    if footer:
        console.print(Align.center(f"({footer})"), highlight=False, style=f'{color} italic')

    console.line()


def print_centered(obj: RenderableType, style: str = '') -> None:
    console.print(Align.center(obj), style=style)


def print_centered_link(url: str, link_text: str, style: str | None = None) -> None:
    print_centered(link_markup(url, link_text, style or ARCHIVE_LINK_COLOR))


def print_color_key() -> None:
    color_table = Table(title=f'Rough Guide to Highlighted Colors', show_header=False)
    num_colors = len(HIGHLIGHTED_GROUP_COLOR_KEYS)
    row_number = 0

    for i in range(0, NUM_COLOR_KEY_COLS):
        color_table.add_column(f"color_col_{i}", justify='center')

    while (row_number * NUM_COLOR_KEY_COLS) < num_colors:
        idx = row_number * NUM_COLOR_KEY_COLS
        color_table.add_row(*HIGHLIGHTED_GROUP_COLOR_KEYS[idx:(idx + NUM_COLOR_KEY_COLS)])
        row_number += 1

    print_centered(vertically_pad(color_table))


def print_header(epstein_files: 'EpsteinFiles') -> None:
    console.print(f"This site isn't optimized for mobile but if you get past the header it should be readable.", style='dim')
    print_page_title(width=TITLE_WIDTH)
    print_other_site_link()
    _print_external_links()
    console.line()
    _print_abbreviations_table()
    epstein_files.print_files_summary()
    print_color_key()
    print_centered(f"if you think there's an attribution error or can deanonymize an {UNKNOWN} contact {CRYPTADAMUS_TWITTER}", 'grey46')
    print_centered('note this site is based on the OCR text provided by Congress which is not always the greatest', 'grey23')
    print_centered(f"(thanks to {link_markup('https://x.com/ImDrinknWyn', '@ImDrinknWyn', 'dodger_blue3')} + others for help attributing redacted emails)")
    print_centered_link(ATTRIBUTIONS_URL, "(some explanations of author attributions)", style='magenta')


def print_json(label: str, obj: object, skip_falsey: bool = False) -> None:
    print(obj)

    if isinstance(obj, dict):
        if skip_falsey:
            obj = {k: v for k, v in obj.items() if v}

        obj = json_safe(obj)

    console.line()
    console.print(Panel(label, expand=False))
    console.print_json(json.dumps(obj, sort_keys=True), indent=4)
    console.line()


def print_numbered_list_of_emailers(_list: list[str | None], epstein_files = None) -> None:
    """Add the first emailed_at timestamp for each emailer if 'epstein_files' provided."""
    current_year = 1990
    current_year_month = current_year * 12
    grey_idx = 0
    console.line()

    for i, name in enumerate(_list):
        indent = '   ' if i < 9 else ('  ' if i < 99 else ' ')
        txt = Text((indent) + F"   {i + 1}. ", style=DEFAULT_NAME_STYLE)

        if epstein_files:
            earliest_email_date = (epstein_files.earliest_email_at(name) or FALLBACK_TIMESTAMP).date()
            year_months = (earliest_email_date.year * 12) + earliest_email_date.month

            # Color year rollovers more brightly
            if current_year != earliest_email_date.year:
                grey_idx = 0
            elif current_year_month != year_months:
                grey_idx = ((current_year_month - 1) % 12) + 1

            current_year_month = year_months
            current_year = earliest_email_date.year
            txt.append(escape(f"[{earliest_email_date}] "), style=f"grey{GREY_NUMBERS[grey_idx]}")

        txt.append(highlighter(name or UNKNOWN))

        if epstein_files:
            num_days_in_converation = epstein_files.email_conversation_length_in_days(name)
            msg = f" ({len(epstein_files.emails_for(name))} emails over {num_days_in_converation:,} days)"
            txt.append(msg, style=f'dim italic')

        console.print(txt)

    console.line()


def print_other_site_link(is_header: bool = True) -> None:
    """Print a link to the emails site if we're building text messages site and vice versa."""
    site_type: SiteType = EMAIL if args.all_emails else TEXT_MESSAGE

    if is_header:
        print_starred_header(f"This is the Epstein {site_type.title()}s site", num_spaces=4, num_stars=14)

    other_site_type: SiteType = TEXT_MESSAGE if site_type == EMAIL else EMAIL
    other_site_msg = "another site for" + (' all of' if other_site_type == EMAIL else '')
    other_site_msg += f" Epstein's {other_site_type}s also generated by this code"
    markup_msg = link_markup(SITE_URLS[other_site_type], other_site_msg, OTHER_SITE_LINK_STYLE)
    print_centered(parenthesize(Text.from_markup(markup_msg)), style='bold')
    word_count_link = link_text_obj(WORD_COUNT_URL, 'site showing the most frequently used words in these communiques', OTHER_SITE_LINK_STYLE)
    print_centered(parenthesize(word_count_link))

    if is_header:
        metadata_link = link_text_obj(JSON_METADATA_URL, 'file metadata containing author attribution explanations', OTHER_SITE_LINK_STYLE)
        print_centered(parenthesize(metadata_link))


def print_page_title(expand: bool = True, width: int | None = None) -> None:
    title_panel = Panel(Text(PAGE_TITLE, justify='center'), expand=expand, style=TITLE_STYLE, width=width)
    console.print(Align.center(vertically_pad(title_panel)))
    print_social_media_links()
    console.line(2)


def print_panel(msg: str, style: str = 'black on white', padding: tuple | None = None, centered: bool = False) -> None:
    _padding: list[int] = list(padding or [0, 0, 0, 0])
    _padding[2] += 1  # Bottom pad
    panel = Panel(Text.from_markup(msg, justify='center'), width=70, style=style)
    actual_padding: tuple[int, int, int, int] = tuple(_padding)

    if centered:
        console.print(Align.center(Padding(panel, actual_padding)))
    else:
        console.print(Padding(panel, actual_padding))


def print_section_header(msg: str, style: str = SECTION_HEADER_STYLE, is_centered: bool = False) -> None:
    panel = Panel(Text(msg, justify='center'), expand=True, padding=(1, 1), style=style)
    panel = Align.center(panel) if is_centered else panel
    console.print(Padding(panel, (3, 0, 1, 0)))


def print_social_media_links() -> None:
    print_centered_link(SUBSTACK_URL, "I Made Epstein's Text Messages Great Again (And You Should Read Them)", style=f'{SUBSTACK_POST_LINK_STYLE} bold')
    print_centered_link(SUBSTACK_URL, SUBSTACK_URL.removeprefix('https://'), style=f'{SUBSTACK_POST_LINK_STYLE} dim')

    social_links = [
        link_text_obj('https://x.com/Cryptadamist/status/1990866804630036988', '@cryptadamist', style=SOCIAL_MEDIA_LINK_STYLE),
        link_text_obj('https://cryptadamus.substack.com/', 'substack', style=SOCIAL_MEDIA_LINK_STYLE),
        link_text_obj('https://universeodon.com/@cryptadamist/115572634993386057', 'mastodon', style=SOCIAL_MEDIA_LINK_STYLE),
    ]

    print_centered(join_texts(social_links, join='     ', encloser='[]'))


def print_starred_header(msg: str, num_stars: int = 7, num_spaces: int = 2, style: str = TITLE_STYLE) -> None:
    stars = '*' * num_stars
    spaces = ' ' * num_spaces
    msg = f"{spaces}{stars} {msg} {stars}{spaces}"
    print_centered(wrap_in_markup_style(msg, style))


def vertically_pad(obj: RenderableType, amount: int = 1) -> Padding:
    return Padding(obj, (amount, 0, amount, 0))


def wrap_in_markup_style(msg: str, style: str | None = None) -> str:
    if style is None or len(style.strip()) == 0:
        return msg

    modifier = ''

    for style_word in style.split():
        if style_word == 'on':
            modifier = style_word
            continue

        style = f"{modifier} {style_word}".strip()
        msg = f"[{style}]{msg}[/{style}]"
        modifier = ''

    return msg


def write_html(output_path: Path) -> None:
    if not args.build:
        logger.warning(f"Not writing HTML because args.build={args.build}.")
        return

    console.save_html(str(output_path), code_format=CONSOLE_HTML_FORMAT, theme=HTML_TERMINAL_THEME)
    logger.warning(f"Wrote {file_size_str(output_path)} to '{output_path}'")


def _print_abbreviations_table() -> None:
    table = Table(title="Abbreviations Used Frequently In These Conversations", header_style="bold", show_header=False)
    table.add_column("Abbreviation", justify="center", style='bold')
    table.add_column("Translation", style="white", justify="center")

    for k, v in HEADER_ABBREVIATIONS.items():
        table.add_row(highlighter(k), v)

    console.print(Align.center(vertically_pad(table)))


def _print_external_links() -> None:
    console.line()
    print_starred_header('External Links', num_stars=0, num_spaces=20, style=f"italic")
    presser_link = link_text_obj(OVERSIGHT_REPUBLICANS_PRESSER_URL, 'Official Oversight Committee Press Release')
    raw_docs_link = join_texts([link_text_obj(RAW_OVERSIGHT_DOCS_GOOGLE_DRIVE_URL, 'raw files', style=f"{ARCHIVE_LINK_COLOR} dim")], encloser='()')
    print_centered(join_texts([presser_link, raw_docs_link]))
    print_centered(link_markup(JMAIL_URL, JMAIL) + " (read His Emails via Gmail interface)")
    print_centered(link_markup(COFFEEZILLA_ARCHIVE_URL, 'Archive Of Epstein Materials') + " (Coffeezilla)")
    print_centered(link_markup(COURIER_NEWSROOM_ARCHIVE_URL, 'Searchable Archive') + " (Courier Newsroom)")
    print_centered(link_markup(EPSTEINIFY_URL) + " (raw document images)")
    print_centered(link_markup(EPSTEIN_WEB_URL) + " (character summaries)")


# if args.deep_debug:
#     print_json('THEME_STYLES', THEME_STYLES)



import json
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

from rich.text import Text

from epstein_files.documents.other_file import OtherFile
from epstein_files.util.constant.strings import JSON


@dataclass
class JsonFile(OtherFile):
    """File containing JSON data."""
    strip_whitespace: ClassVar[bool] = False

    def __post_init__(self):
        super().__post_init__()

        if self.url_slug.endswith('.txt') or self.url_slug.endswith('.json'):
            self.url_slug = Path(self.url_slug).stem

        self._set_computed_fields(text=self.formatted_json())

    def category(self) -> str:
        return JSON

    def formatted_json(self) -> str:
        return json.dumps(self.json_data(), indent=4)

    def info_txt(self) -> Text | None:
        return Text(f"JSON file, possibly iMessage or similar app metadata", style='white dim italic')

    def is_interesting(self):
        return False

    def json_data(self) -> object:
        with open(self.file_path, encoding='utf-8-sig') as f:
            return json.load(f)
